# __init__.py  


#  checks if .exr  has been installed
#  if found (and is latest version) launches main.register()
#  if not found, then tries to copy instance over
#  if works then launches main.register()
#  if not then invariably will be caused by permission error
#  does popup to warn user, and requests they restart blender with admin or root permissions
#  when they acknowledge (click ok)  then blender shuts down immediately,
#  -- without even starting to attempt loading or installing rest of app
#  acts as 'gatekeeper'


import bpy  
import sys
import os
import shutil

module_path = os.path.dirname(__file__)     # make sure python can find the pristina modules  
if module_path not in sys.path:             
    sys.path.append( module_path)           # -- update sys.path if necessary     




# ******** must set version both here below and in nested_ui_panels.py  **************
# now set in 3 places, init and nested panels  and main

bl_info = {
    "name":        "Pristina (TM)    (c)2025 Ideal Real Labs",      # todo  change all (tm) marks to (r) when got
    "version":     (0, 46),             
    "blender":     (2, 80, 0),
    "category":    "Render",
    "description": "Supercharge your archviz workflow with Pristina AI"
                   " -- turn any image into a professional-grade image",
    "warning":     "*** Installation requires elevated privileges (admin or root) ***",
    "wiki_url":    "https://pristina.ai",      
}


vtup                = bl_info['version']            # get tuple from above bl info spec
version             = f"v{vtup[0]}.{vtup[1]}"       # convert to string eg  'v0.41'



launch_popup     = False                    # (if permission error) popup notifies user of the need for elevated privileges
                                            # and asks user to restart as admin or root user
                                            # flag has 'one shot' use -- when True it is acted on once, 
                                            # and immediately set back False to stop repeated triggering of popup
     


    
                                            # get visibility for the package and modules --- append sys path
plugin_dir   = os.path.dirname(__file__)
if plugin_dir not in sys.path:  sys.path.append( plugin_dir)

import EULA_agreement_popup as eula


from logger_module    import   Logger       # setup logger
logger       = Logger.get_logger()      
logger.info( f'plugin_dir ={plugin_dir}  plugin version = {version}')    


#from main  import   version_in_main         # set this in main so that 
#version_in_main = version                   # log uploader initialization can use it
   
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# Blender's installation directory
python_executable = sys.executable
blender_install_dir = os.path.dirname(os.path.dirname(python_executable))
logger.info(f"^^^^^^^^^^^^^^^^^^^^^^^  Blender installation directory: {blender_install_dir}")
# eg  Blender installation directory: C:\Program Files\Blender Foundation\Blender 4.1\4.1\python
# r"C:\Program Files\Blender Foundation\Blender 4.1\4.1\python\lib\site-packages"   # example path to supabase package
sp = os.path.join( blender_install_dir, "lib", "site-packages")
sys.path.append( sp)
logger.info(f"^^^^^^^^^^^^^^^^^^^^^^^  Blender site packages: {sp}")

version_path = os.path.dirname(blender_install_dir)
exr_expected_at = os.path.join( version_path, 'datafiles', 'studiolights', 'matcap', 'check_normal_pristina.exr')
logger.debug(f'exr expected at {exr_expected_at}')
found = os.path.exists( exr_expected_at)
if found:
    logger.debug(f'exr found in place ok')
else:
    logger.critical(f'\n\n\n\t\t************* exr not found ***************\n\n\n')
    

#!!!!!!!!!!!!!!!!!!!!!!!!!!!


logger.info(f"^^^^^^^^^^^^^^^^^^^^^^^  Blender installation directory: {os.path.dirname(os.path.dirname(sys.executable))}")








            
def ensure_exr_file():

    global launch_popup

    try:
        source, dest  = eula.get_exr_source_and_dest()
   

        # ----------------------------------------------------
        # ***  Check if destination file exists already, and has the same size
        # ***  (if so, job already done) --  [additional datetime check was found not to be reliable !]
        
        # todo -- when making a new .exr file, just ensure its size is new (at least not same as previous, ideally completely new), 
        # so this method can tell it is different from previously installed
        
        if os.path.exists( dest):
            source_stat = os.stat( source)
            dest_stat   = os.stat( dest)
            
            if (source_stat.st_size  == dest_stat.st_size ):       
                logger.info(f"   .exr file is already in place and is up-to-date: \n{dest}")
                eula.launch_main_app()    # don't mind redundancy in copying exr over again, just easier.
                return      
        # ----------------------------------------------------
        

        # no .exr in place, try and copy it over, exception if permission not sufficient
        
    
        shutil.copy2(source, dest)      # shutil.copy2 preserves file metadata, trips out here if lacking permissions, misses EULA
                                        # exception is caught and popup is launched to warn user of permissions needed
                                        # first time here do only as test for permission
                                        
        logger.info( f'test copy proved we have permission')                                        
                                        
        os.remove(dest)                 # remove test copy now, so if eula not agreed, then already cleaned up 
                                        # redo in main install
        
        logger.info( f'test copy file removed')                                        
        
        
        
        # MUST NOW GET THE 'EULA' AGREED             
        
        logger.info(f"about to register eula")
        eula.register()
        logger.info(f"eula registered. This launches modal EULA popup")        
        logger.info(f'\n\n\t********* EULA has returned, even though button not pressed\n\n')
        logger.info(f'\n\n\t*** returning from ensure_exe_file() method !! nb not yet copied over, done in launch_main_app() \n\n')
        return             
        
       
       
    except PermissionError as ex:
        logger.warning(f'\n\n\t*** Pristina install requires elevated privileges (admin or root) to avoid permission errors ***\n\n')
        logger.info(f'\n\n\t***** Now setting flag to do popup for user, then quit Blender *****\n\n')
        launch_popup     = True
        check_permission()    # start the process for notfifying user of requirement and seeking confirmation for shutdown
        return

    
    except Exception as ex:
        msg = f'copy exr unexpected exception: {ex}'
        logger.exception( msg)     
        raise RuntimeError( msg)  #  this is then a dev error, not for the user
            
   


# ----------------------------------------------------------------------------------    popup code to deal with user trying to install 
#                                                                                       without having elevated permissions
 
def check_permission():
    logger.info(f'\n\n\t*****permission error flag={launch_popup} *****\n\n')
    
    if launch_popup:
        if "ConfirmShutdownOperator" not in bpy.types.__dict__:
            try:
                bpy.utils.register_class(ConfirmShutdownOperator)
            except Exception as ex:
                logger.warning(f'could not register ConfirmShutdownOperator')                                    
        
        logger.info("Permission error detected, flag set to True.")
        bpy.app.timers.register(trigger_confirmation_dialog, first_interval=0.1)




class ConfirmShutdownOperator(bpy.types.Operator):          # warn user to restart Blender as admin or root
    
    bl_idname  = "wm.confirm_shutdown_dialog"
    bl_label   = "To install: restart Blender as Admin or Root"    # title of the popup dialog
    bl_options = {'REGISTER', 'UNDO'}
    
    def execute(self, context):
        logger.info("ConfirmShutdownOperator.execute() called, cleaning up and shutting down Blender...")        
        eula.cleanup("Pristina_plugin")
                      
        try:    
            unregister_after_permission_error()             # much reduced unregister method, just for situation         
        except Exception as ex:                             # where permission error is detected and want clean shutdown
            logger.exception( f'{ex}')                      # this happens ahead of all the normal registrations, when .exr installed ok
                          
        bpy.ops.wm.quit_blender()                           # Quit Blender
        return {'FINISHED'}                                 # never reach? Don't mind this
        

    def cancel(self, context):                              # Treat cancel button press same as the OK by calling execute        
        logger.info("ConfirmShutdownOperator.cancel() called, treating cancel as OK...")
        return self.execute(context)
            
    
    def draw(self, context):
        layout = self.layout
        layout.label(text="(Elevated privileges are only", icon='ERROR')    # notification text shown in box
        layout.label(text="          required during installation)")        
        layout.label(text="Press OK  (or Cancel)  to quit Blender.")        
    
    
    def invoke(self, context, event):
        logger.info("ConfirmShutdownOperator.invoke() called, showing confirmation dialog...")               
        return context.window_manager.invoke_props_dialog(self)  # Open a modal dialog with only OK and Cancel buttons




def trigger_confirmation_dialog():  # Function to trigger the modal dialog safely in the Preferences window context
    global launch_popup
    
    if launch_popup:
        launch_popup = False        # Reset the flag immediately to avoid repeated triggering
        
        logger.info("Diagnostic: Invoking custom confirmation dialog in Preferences context...")
        
        preferences_area = None     # Make sure the dialog is associated with the Preferences window context

        # Find the Preferences window context
        for window in bpy.context.window_manager.windows:
            for area in window.screen.areas:
                if area.type == 'PREFERENCES':
                    preferences_area = area
                    break

        if preferences_area:
            with bpy.context.temp_override(window=window, area=preferences_area):
                bpy.ops.wm.confirm_shutdown_dialog('INVOKE_DEFAULT')
        else:
            logger.info("Diagnostic: Preferences area not found, cannot trigger confirmation dialog.")
    
    return None                     # Stops the timer from repeating


                


# -------------------------------------------------------------------------------------  dummy operator


class DummyOperator(bpy.types.Operator):    # Register this dummy operator to avoid 'no register()' error
                                            # as real registrations are made later, conditionally
                                            # may need to have register and unreg that blender uses, but these 
                                            # work on a dynamic list, initially dummy only and then full set for plugin
    bl_idname = "wm.dummy_operator"
    bl_label  = "Dummy Operator"

    def execute(self, context):
        return {'FINISHED'}

# --------------------------------------------------------------------------------------



def register():
    if "DummyOperator" not in bpy.types.__dict__:        
        try:    
            bpy.utils.register_class( DummyOperator)     # may already have been registered, if user reg'd and unreg'd and reg'd again                                            
        except Exception:
            pass       
        
    logger.info( f'calling ensure_exr_file()')
    ensure_exr_file()                       # if ensure is happy with .exr file, then it will launch the app via main.py module







def unregister_after_permission_error():    # only need a couple of unregs here, as we made sure to quit before getting into main.py setups
    if "DummyOperator" in bpy.types.__dict__:
        try:    
            bpy.utils.unregister_class( DummyOperator)     # may already have been registered, if user reg'd and unreg'd and reg'd again                                            
        except Exception:
            pass       
        
    if "ConfirmShutdownOperator" in bpy.types.__dict__:        
        try:    
            bpy.utils.unregister_class(ConfirmShutdownOperator)                                    
        except Exception:
            pass       
    
    
    
    
def unregister():                           # use the normal unreg  in main.py module
    import main
    main.unregister()
