# cryptomatte_renderer.py


import bpy
import os

from logger_module                  import   Logger

logger       = Logger.get_logger()      ####################  global, is set up on load 




class CryptomatteRenderer:
    def __init__(self, object_list, output_folder):
        self.object_list      = object_list
        self.output_folder    = output_folder
        self.file_output_path = output_folder
        self.file_name        = "filename"

        # Switching to Cycles render engine
        bpy.context.scene.render.engine = 'CYCLES'

        # Enable nodes in the compositor
        bpy.context.scene.use_nodes = True
        self.tree                   = bpy.context.scene.node_tree

        # List of potential view layer names
        view_layer_names      = ["ViewLayer", "View Layer"]

        # Flag to track if a view layer has been found
        view_layer_found      = False
        
        logger.info( "RUNNING CRYPTOMATTE")
        # Iterate through possible view layer names
        for name in view_layer_names:
            if name in bpy.context.scene.view_layers:
                # Get the view layer
                view_layer    = bpy.context.scene.view_layers[name]
                
                # Enable the Cryptomatte object pass
                view_layer.use_pass_cryptomatte_object = True
                # Enable the Cryptomatte material pass
                view_layer.use_pass_cryptomatte_material = True
                # Enable the Cryptomatte asset pass
                view_layer.use_pass_cryptomatte_asset = True

                logger.info( f"Cryptomatte for object, material, and asset enabled for '{name}'.")
                view_layer_found = True
                break  # Exit the loop once a match is found

        if not view_layer_found:
            logger.warning( f"None of the specified view layers found in the scene.")

        # Remember the initial node count
        self.initial_node_count = len(self.tree.nodes)

        # Create a single Render Layers node
        self.render_layers_node = self.tree.nodes.new(type='CompositorNodeRLayers')

        # Create Cryptomatte nodes based on the object list
        self.create_cryptomatte_setup()
        
        self.render_and_cleanup()

    def create_cryptomatte_setup(self):
        for index, object_name in enumerate(self.object_list):

            logger.info( f"\n\n\tLIST ITEM {index}:  {object_name}")

             # Add a new Cryptomatte node for each object
            cryptomatte_node           = self.tree.nodes.new(type='CompositorNodeCryptomatteV2')
            cryptomatte_node.matte_id  = object_name

            # Create a File Output node for each object
            file_output_node           = self.tree.nodes.new(type='CompositorNodeOutputFile')

            # Set the output filename and connect nodes
            filename                   = object_name + "_mask"
            file_output_node.base_path = self.output_folder  # Set your desired output folder
            file_output_node.file_slots[0].path = filename

            # Connect nodes
            links                      = self.tree.links.new
            links(self.render_layers_node.outputs["Image"], cryptomatte_node.inputs[0])
            links(cryptomatte_node.outputs["Matte"], file_output_node.inputs["Image"])


    def save_render_result(self, filepath):
        try:
            bpy.context.scene.render.filepath = filepath
            bpy.ops.render.render(write_still=True)
            logger.info( "Render saved successfully.")
        except Exception as e:
            logger.exception( f"Error saving render: {str(e)}")
            raise
        finally:
            bpy.context.scene.render.filepath = ""  # Reset the filepath to avoid the mentioned error
            

    def render_and_cleanup(self):
        logger.info( "RENDER AND CLEANUP STARTED")
        filepath_render_output = os.path.join(self.file_output_path, f'{self.file_name}.png')
        result                 = ('FINISHED', "Initial value")  # Placeholder for result message
        
        try:
            self.save_render_result(filepath_render_output)
            logger.info( f"Render saved successfully.  {filepath_render_output}")
            result             = ('FINISHED', f"Rendering Pristina to {filepath_render_output}")
        except Exception as e:
            logger.exception( f"Error saving render: {e}")
            result             = ('CANCELLED', f"Error saving render: {e}")
        finally:
            # Cleanup code now in the finally block to ensure it always runs
            logger.info( "BEGIN CLEANUP")
            for node in list(self.tree.nodes)[self.initial_node_count:]:  # Added list() to copy the list, ensuring stable iteration and modification
                self.tree.nodes.remove(node)

            logger.info( "Blender script executed successfully.")
        
        return result


