# LogUploader.py


# ******* THIS MODULE IS REUSED BY OTHER PRISTINA APPS  (eg 3ds and Blender plugins, and office and pod task processors ***

# ******* ONLY MAKE CHANGES IF ESSENTIAL, AND THEN TEST ON ALL USERS.
# ******* IDEALLY KEEP FIXED

# THE ONLY CUSTOMIZATION REQUIRED FOR THIS MODULE IS TO SUPPLY A CALLBACK REF, WHICH CAN THEN BE USED TO
# GET THE APP NAME OF THE CURRENT USER  (eg blender_plugin)
# For example see blender plugin __init__()  where a callback is defined and then registered --- very simple approach


import os
import time
import zipfile
import threading
from datetime import datetime
from logger_module import Logger
from sessions import SessionManager


# **************  PLUGIN VERSION  ************************


'''
This uploader has similar an RLS permission to that for auth user uploading job to job bucket
In this case permission to INSERT will be into the 'pristina_logs' bucket. Supabase SQL for this is:


-- POLICY user allow INSERT logs

DO $$
BEGIN
    -- Check if the policy already exists
    IF EXISTS (
        SELECT * 
        FROM pg_policies 
        WHERE policyname = 'POLICY user allow INSERT logs'
        AND tablename    = 'objects'
        AND schemaname   = 'storage'
    ) THEN
        -- If the policy exists, drop it
        DROP POLICY "POLICY user allow INSERT logs" ON storage.objects;
    END IF;

    -- Now create the updated policy
    CREATE POLICY "POLICY user allow INSERT logs"
    ON storage.objects
    FOR INSERT
    TO authenticated
    WITH CHECK (
        bucket_id = 'pristina_logs'
        AND auth.role() = 'authenticated'
    );
END $$;

'''

logger = Logger.get_logger()                # Global logger setup


class LogUploader:
    """
    Handles zipping and uploading log files to Supabase
    good:  called from file_utils module -- download_and_unzip()
    fail:  called from fifo_queue module -- check_fifo_queue()
    """
    
    callback_for_distro_and_version = None  # get the distro and version strings from the code that is calling/using this class
                                            # could be 3ds or Blender plugin or office or pod task processor, 
                                            # then in plugin or tp, do: LogUploader.callback_for_distro_and_version = distro_callback
       

    LOGS_UPLOAD_NONE        = 0             # available modes of functioning
    LOGS_UPLOAD_ALL         = 1
    LOGS_UPLOAD_FAILS_ONLY  = 2

    mode = LOGS_UPLOAD_ALL                  # current mode setting from above <<<<<<<<<<<<< SET MODE HERE

    lock = threading.Lock()                 # Class-level lock for thread safety
    

    @classmethod
    def on_good_success(cls, task_id):
        """
        good: called from file_utils module -- download_and_unzip()
        When a job finishes ok: (ie when goes to FETCHED status, upload log file so far -- may include prev jobs in this run)
        """
        result  = 'GOOD'
        logfile = Logger._logfile
        
        if cls.mode == cls.LOGS_UPLOAD_ALL:       
            
            thread  = threading.Thread(     # Spawn a daemon thread for the upload
            
                                target  = cls.upload,
                                args    = (result, logfile, task_id),
                                daemon  = True )
            thread.start()
            return 'done'
        else:
            return 'skipped'
            
            

    @classmethod
    def on_job_fail(cls, task_id):
        """
        fail: called from fifo_queue module -- check_fifo_queue()
        When a job fails, still upload log file so far -- which may include prev jobs in this run
        """
        result  = 'FAIL'
        logfile = Logger._logfile
        
        if cls.mode in [cls.LOGS_UPLOAD_ALL, cls.LOGS_UPLOAD_FAILS_ONLY]:
            
            thread  = threading.Thread(     # Spawn a daemon thread for the upload
            
                                target  = cls.upload,
                                args    = (result, logfile, task_id),
                                daemon  = True )
            thread.start()
            return 'done'
        else:
            return 'skipped'
            
            

    @classmethod
    def upload(cls, result, logfile_path, task_id):
        """
        Zips a logfile and uploads it to Supabase under the specified path.

        Args:
            logfile_path (str): Path to the logfile.
            task_id (str): Task ID to include in the upload path.
            version (str): Plugin version.
        """
        with cls.lock:                                  # Ensure the method is executed by only one thread at a time
            
            try:
                distro, version = cls.callback_for_distro_and_version()     # calling code (eg plugin or processor) 
                                                                        # will supply its own name and app version
            except:
                msg = "callback must be registered during program startup"
                logger.error( msg)
                raise RuntimeError( msg)


            logger.info(f'version is {version}')

            if not os.path.exists(logfile_path):
                logger.error(f"Logfile {logfile_path} does not exist.")
                return None

            zip_filename = f"{result}_{distro}_{version}_log_{task_id}.zip"  # Prepare zip file
            zip_filepath = os.path.join(os.path.dirname(logfile_path), zip_filename)

            try:  # Create the zip file
                storage_path = f"pristina_logs/task_id_{task_id}/{zip_filename}"
                logger.debug(f"Attempting to zip logfile and upload to\n\t\tSupabase storage path: {storage_path}\n\n"
                             f"{'-' * 120}\n###\n")

                with zipfile.ZipFile(zip_filepath, 'w', zipfile.ZIP_DEFLATED) as zipf:
                    zipf.write(logfile_path, os.path.basename(logfile_path))

                # Ensure the zip file exists and is not empty
                if not os.path.exists(zip_filepath) or os.path.getsize(zip_filepath) == 0:
                    logger.error("Verification failed: Zip file was not created or is empty.")
                    return None

            except Exception as ex:
                logger.error(f"Error zipping logfile: {ex}")
                return None

            thread_id = SessionManager.get_thread_id_with_name('UploadLog')
            logger.debug(f'got thread_id={thread_id}')

            try:
                sb_client = SessionManager.get_client(thread_id)

                with open(zip_filepath, "rb") as zip_file:
                    zip_content = zip_file.read()

                # Upload to Supabase
                upload_result = sb_client.storage.from_("pristina_logs").upload(storage_path, zip_content)
                logger.debug(f"Upload result: {upload_result}")
                logger.debug(f"Successfully uploaded logfile to {storage_path}")
                return storage_path

            except Exception as ex:
                logger.warning(f"Error uploading logfile: {ex}")   # during testing see duplicate resource error when repeating 
                # not a show stopper, just log and carry on
                return None

            finally:
                SessionManager.logout(thread_id)            # specific to 'UploadLog' session
                if os.path.exists(zip_filepath):            # Clean up
                    os.remove(zip_filepath)


# ----------------------------------------------------------------------------  test/demo code only


if __name__ == "__main__":
    import random

    # Ensure Supabase session is set up
    if SessionManager.email is None:
        SessionManager.setup('neal@irlabs.co.uk', '123456789')

    good_task_id = f"{random.randint(1000, 9999)}"    # random is ok, as we are only going to bucket, not task_queue table id
    LogUploader.on_good_success(good_task_id)

    fail_task_id = f"{random.randint(1000, 9999)}"
    LogUploader.on_job_fail(fail_task_id)

    # Allow threads to complete
    time.sleep(5)

    # Clean up sessions
    SessionManager.close_all_sessions()
