# logger_module.py


import logging
import os
import sys
import time
import traceback
from threading import Lock


# writes logging data to console  (info  level and up), 
#                     and to file (debug level and up).

# log file is to be found at "..\Project_Pristina\Tests\pristina.log"

# for example, at    r"C:\Users\cex\deep\FINZ\ipynb\Project_Pristina\Tests\pristina.log"
            

'''
Standard Logging Levels
NOTSET - 0
This level is typically used to initialize the level attribute in logging objects and has a special behavior when used in libraries. Loggers with a level of NOTSET will defer to their parent's level.


DEBUG - 10
This level is used for detailed diagnostic information, used primarily for problem diagnosis. It captures a high level of detail mainly interesting during development and for troubleshooting.


INFO - 20
This level is used for general interest messages which confirm that things are working as expected. In a production environment, this is usually the lowest level you would log at.


WARNING - 30
This level indicates something unexpected or a problem in the near future (e.g., 'disk space low'). The software is still working as expected.


ERROR - 40
This level indicates a more serious problem, due to which the software has not been able to perform some function.


CRITICAL - 50
This level indicates a serious error, indicating that the program itself may be unable to continue running.

'''




### https://stackoverflow.com/questions/6729268/log-messages-appearing-twice-with-python-logging    duplication of messages


class Logger:                            # singleton pattern, one logger for whole application

    current_file_path = os.path.abspath( __file__ )            
    #print(f'current_file_path={current_file_path}')
    
    current_dir       = os.path.dirname( current_file_path )
    #print(f'current_dir={current_dir}')
    
    parent_dir        = os.path.dirname( current_dir )
    # print(f'parent_dir={parent_dir}')
    
    pparent_dir        = os.path.dirname( parent_dir )
    # print(f'pparent_dir={pparent_dir}')
    
    ppparent_dir        = os.path.dirname( pparent_dir ) 
                                        # eg pparent_dir=C:\Users\cex\AppData\Roaming\Blender Foundation\Blender\4.1\scripts
    # print(f'ppparent_dir={ppparent_dir}')
    
    project_root      = os.path.dirname( current_dir )
    #print(f'project_root={project_root}')
        
    

    _file_level_fixed = logging.DEBUG   # the log file written to disk has full detail of logs, 
                                        # only 'info' level and above are written to console

    _logger           = None
                                        # TODO DECIDE ABOUT LOGFILE IN POD (USE AT ALL? WHERE PERMISSION TO WRITE TO? HOW TO ACCESS?)
                                        
    # _logfile          = os.path.join( r'C:\Users\cex\Downloads', 'pristina.log')    # log file in SAME FOLDER AS MODULES
    # _logfile          = os.path.join( parent_dir, 'pristina.log')    # log file in SAME FOLDER AS MODULES
    # _logfile          = os.path.join( current_dir, 'pristina.log')    # log file in SAME FOLDER AS MODULES
    _logfile          = os.path.join( pparent_dir, 'pristina.log')    # log file in SAME FOLDER AS MODULES
    
    _level_stack      = []    # Stack to keep track of levels just for the console, file level is fixed
    _console_handler  = None            # Console handler as a class variable
    _instance         = None
    _lock             = Lock()          # Thread-safe singleton pattern
    file_handler      = None
  
  
  
    @classmethod
    def manage_prev_logfiles(cls):
        with cls.lock:
            logfile = Logger._logfile
            logdir = os.path.dirname(logfile)

            prev = os.path.join(logdir, "pristina_prev.log")
            if os.path.exists(prev):
                os.remove(prev)

            curr = os.path.join(logdir, "pristina.log")
            if os.path.exists(curr):
                Logger.clear()  # Clear the logger to release the logfile
                time.sleep(0.1)  # Allow time for file handle release

                retries = 20  # Retry mechanism for renaming
                while retries > 0:
                    try:
                        os.rename(curr, prev)
                        break
                    except PermissionError:
                        time.sleep(0.1)  # Wait before retrying
                        retries -= 1

                if retries == 0:
                    raise RuntimeError(f"Failed to rename log file {curr} to {prev}. File may still be in use.")
  
  
  

    def __new__(cls):
        with cls._lock:
            if cls._instance is None:
            
                print(f'\n\nnew logger instance, logfile to be found at\n\n\t\t{ cls._logfile}\n\n\n')
            
                try:
                    
 
                    # ensure folder for log file
                    log_folder = os.path.dirname( Logger._logfile)
                    os.makedirs( log_folder, exist_ok=True)
                                                    
                
                    # Clearing root logger handlers
                    for handler in logging.root.handlers[:]:
                        logging.root.removeHandler(handler)
                    logging.root.setLevel(logging.CRITICAL)  # Set root logger to critical

                    cls._instance = super(Logger, cls).__new__(cls)

                    # Configure the logger
                    cls._instance._logger = logging.getLogger('Pristina_logger')
                    cls._instance._logger.setLevel(logging.DEBUG)
                    cls._instance._logger.propagate = True
                    
                    

                    # Decide file output format
                    format = "\n%(asctime)s - %(levelname)-8s - %(module)s - line %(lineno)3d in %(funcName)s - %(message)s"
                    formatter = logging.Formatter(format, datefmt='%H:%M:%S')  # time of day only

                    # Create file handler
                    
                    file_handler = logging.FileHandler(cls._logfile, mode='w')  # 'w' for overwrite, 'a' for append
                    file_handler.setLevel( cls._file_level_fixed)  # not affected by console pushes and pops
                    file_handler.setFormatter(formatter)
                    logging.root.addHandler( file_handler)                
                        
                    

                    # Console handler configuration
                    console_format = "\n%(asctime)s - %(levelname)-8s - %(module)s - line %(lineno)3d in %(funcName)s - %(message)s"
                    console_formatter = logging.Formatter(console_format, datefmt='%H:%M:%S')
                    console_handler = logging.StreamHandler()
                    cls._console_handler = console_handler
                    #console_handler.setLevel(logging.INFO)
                    console_handler.setLevel(logging.INFO)
                    console_handler.setFormatter(console_formatter)
                    logging.root.addHandler(console_handler)

                    # Initialize level stack with default console level
                    cls._level_stack.append(console_handler.level)

                    return cls._instance
                    
                except Exception as ex:
                    print(f'\n\nException in logger setup:  {str(ex)}\n\n')

 
    @classmethod
    def get_logger(cls):
        # print('&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& getting logger &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&')
        if cls._instance is None:
            cls()
        return cls._instance._logger
    
    
    
    '''
    @classmethod
    def push_level(cls, level):
        print(f'\n\n\nin push_level() before: level={cls._console_handler.level}\n\n\n')
    
        if cls._console_handler:
            cls._level_stack.append(level)
            print(f'Setting console level to {level}')
            cls._console_handler.setLevel(level)
            print(f'\n\n\nin push_level() after: level={cls._console_handler.level}\n\n\n')
        else:
            print("No console handler found.")

    @classmethod
    def pop_level(cls):
        if len(cls._level_stack) > 1:  # Ensure default level remains
            cls._level_stack.pop()
        if cls._console_handler:
            # print('Resetting console level')
            cls._console_handler.setLevel(cls._level_stack[-1])
            
    '''
            

    @classmethod
    def clear(cls):
        """
        Clears the logger instance and releases its resources.
        Includes debug prints for troubleshooting.
        """
        logfile = cls._logfile
        print(f"\n[DEBUG] clear() called. Target logfile: {logfile}")

        if cls._instance and cls._instance._logger:
            logger = cls._instance._logger
            handlers = logger.handlers[:]
            print(f"[DEBUG] Found {len(handlers)} handler(s) attached to the logger.")

            for i, handler in enumerate(handlers):
                print(f"[DEBUG] Closing handler {i}: {handler}")
                try:
                    handler.close()  # Close the handler to release the file handle
                    logger.removeHandler(handler)  # Remove the handler from the logger
                    print(f"[DEBUG] Handler {i} closed and removed successfully.")
                except Exception as e:
                    print(f"[ERROR] Exception while closing handler {i}: {e}")

            print("[DEBUG] Shutting down logging subsystem.")
            logging.shutdown()

            cls._console_handler = None
            cls._instance._logger = None
            cls._instance = None
            print("[DEBUG] Logger instance and handlers cleared.")

        else:
            print("[DEBUG] No active logger instance found.")

        # Wait to allow file handle release
        time.sleep(0.2)
        print("[DEBUG] Delay added to ensure file handle release.")

        # Verify the file's status
        if os.path.exists(logfile):
            try:
                with open(logfile, "a") as f:
                    print(f"[DEBUG] Successfully opened {logfile} for appending. File is not locked.")
            except Exception as e:
                print(f"[WARNING] Cannot open {logfile}. File may still be locked: {e}")
        else:
            print(f"[DEBUG] Logfile {logfile} does not exist.")
            
            
            

    @classmethod
    def XXXclear(cls):    # was not releasing log file quick enough, so could not be removed: permission error
        if cls._logger:
            for handler in cls._logger.handlers[:]:  # Iterate over a copy of the list
                handler.close()
                cls._logger.removeHandler(handler)  # Corrected to use cls._logger

        cls._logger          = None  # Reset the logger
        cls._console_handler = None  # Also reset console handler
        
        
    @classmethod
    def YYYclear(cls):
        """
        Clears the logger instance and releases its resources.
        """
        if cls._instance and cls._instance._logger:
            logger = cls._instance._logger
            handlers = logger.handlers[:]

            for handler in handlers:
                handler.close()               # Close the handler to release file handles
                logger.removeHandler(handler) # Remove the handler from the logger

            cls._console_handler = None       # Reset console handler reference
            cls._instance._logger = None      # Reset logger reference
            cls._instance = None              # Reset the singleton instance
        
            
     


    @classmethod
    def print_log( cls):
        print(f'\n\nNow printing out the log file ({cls._logfile})\n')
        with open( cls._logfile, 'r') as file:
            for line in file:
                print( line, end='')
        print('\n\n')


    @classmethod
    def handle_exception( cls, ex):       
        mm, f, n =  cls.get_exception_message( ex)   
        # open_notepadPP_file_atline_number( f, n)        # use file and line number to open position in notepad++
        cls._logger.error(f'\n\nIn handle_exception() line={n},  f={f}\nmm message=\n<<{mm}>>\n\n')
        status = 1    # 0 for OK, 1 for error
        cls.print_log()
        print(f'Exiting app now')
        sys.exit( status) ###################################### EXIT HERE
        #return status


    @classmethod
    def get_current_working_directory( cls):                # allows dev runs calling modules in pc folder(s),  
        return os.path.abspath('.')                     # then make and run plugin in Blender with modules in plugin zip 






def open_notepadPP_file_atline_number( f, n):      
    # example interfacing to editor, 
    # similarly sublime                   subl file_path:line_number --command focus
    if os.path.isfile( f):
        print(f'os path says  {f} is a file ')
        path = r"C:\Program Files\Notepad++\notepad++.exe"
        command = f'"{path}" -n{n}  "{f}"'
        # command = f'"{path}" -n{line_number} -c{column_number} "{file_path}"'
        subprocess.run(command)


#def handle_exception( ex):
    #print( f"An error occurred: {str(e)}" )
        # Optionally, re-raise the exception if you want it to propagate:
        #raise





# --------------------------------------------------------------------- demo code
# in prompt cd  'C:\\Users\\cex\\deep\\FINZ\\ipynb\\project_pristina\\src'
# then      python logger_module.py

def demo_logger():

    Logger.clear()
    logger = Logger.get_logger()
    print('\n\n-----------------------------------------------------------------------------\n\n')
    print('Logger demo first printing to console, then reading back log file\n')

    Logger.push_level(logging.INFO)
    logger.debug('This message should not show in the console, temporary level setting')
    Logger.pop_level()

    logger.debug('This message should show in console')
    logger.info('So should this -- major headings in progress of program')
    logger.warning('And this, too -- eg resources are low, or took longer than expected, etc')
    logger.error('And non-ASCII Malmö, --- eg retry if say file not found')
    logger.critical('Cannot proceed further, better try and shut down gracefully')
    
    try:
        a = 1/0
    except Exception as ex:
        Logger.handle_exception( ex)
    
    

    time.sleep(1)
    Logger.print_log()
    
    

if __name__ == "__main__":
    demo_logger()
