import bpy
import os
import shutil
import sys
import time


# Crucial not to re-arrange the order of these methods and imports, done this
# way to allow upfront checking for .exr file present, or not, 
# in which case must have root or admin permissions to copy it in.
# If found not to have permissions, then
# idea is to notify user, get confirmation ok, and shutdown Blender as soon as 
# possible to avoid all other checks and imports and starting the install proper
# then there should be nothing to clean up (as would be after failed install)


# ---------------------------  cannot use sleep as would pause all Blender threads
# used by checking for .exr status while True loop to delay next check

def wait_n_seconds( n ):
    print(f"Waiting for {n} seconds")
    # This function will be called after 'n' seconds
    bpy.app.timers.register(after_wait, first_interval=n)

def after_wait():
    print("Finished waiting")
    # Continue with your script or logic here
    return None  # Stop the timer

# usage:  wait_n_seconds( 1 )
# --------------------------------------------------------------------------------



########## TELL SYS.PATH WHERE TO FIND THE PRISTINA MODULES ##########

module_path = os.path.dirname(__file__)

if module_path in sys.path:                 # Remove if it is already in sys.path, as may not be top
    sys.path.remove( module_path)
    
if module_path not in sys.path:                    
    sys.path.insert( 0, module_path)        

    
########## SET UP LOGGING ##########    
    
from logger_module                  import   Logger
logger       = Logger.get_logger()      ####################  global, is set up on load 
logger.info('\n\n\n\n____________________________________________________________________________________________________')
logger.info(f'module path = \n\t{module_path}')
logger.info( f'module_path now first in sys path list = { module_path == sys.path[0]}')  # check top of list



########## IMPORTS ##########   --- see also in register(), supabase related

from pip_installs                   import   *

ensure_dependencies_installed()     # in pip_installs -- installs supabase and its dependencies (about 25 packages in all)


'''  toml not needed now as blender manifest is standalone, and not to be used for any custom global config vars etc.

if sys.version_info >= (3,11):      # TOML / TOMLLIB     NB must compare versions as tuples, not decimal point as then float      
    import tomllib  as toml         # tomllib is built into python 3.11 and upwards
else:
    import toml                     # (python 3.10 and below) -- toml external module will have been pip installed in 'pip_installs.py'
'''


# these depend on supabase, so could not import until supabase in place ok:
    
from sessions                       import   SessionManager  
from fifo_queue_module              import   *     

from ot_login                       import   PRISTINA_OT_Login

from pristina_item_and_properties   import   PristinaItem
from pristina_item_and_properties   import   PristinaProperties
from ot_render_pristina             import   PRISTINA_OT_RenderPristina 
from ot_add_move_remove_items       import   PRISTINA_OT_AddItem
from ot_add_move_remove_items       import   PRISTINA_OT_MoveItem
from ot_add_move_remove_items       import   PRISTINA_OT_RemoveItem
from ot_select_object               import   PRISTINA_OT_SelectObject

from nested_ui_panels               import   PRISTINA_PT_MainPanel
from nested_ui_panels               import   PRISTINA_PT_GlobalSettings
from nested_ui_panels               import   PRISTINA_PT_ObjectSettings
from nested_ui_panels               import   PRISTINA_PT_AccountSettings

from ot_sundry                      import   PRISTINA_OT_OpenLink
from ot_sundry                      import   PRISTINA_OT_SortByDistance
from ot_sundry                      import   ERROR_OT_popup_message
from ot_sundry                      import   LoadImageFromFileOperator  # Import the new operator

from image_resizer                  import   ImageResizer   # Import the new module

from compositor_setup               import   CompositorSetup

from task_manager                   import   TaskManager    # just for shutdown in unregister()

from popup_for_job_fail             import   TaskFailedPopupOperator
from popup_for_low_credits          import   CreditsLowPopupOperator
from popup_for_bad_login            import   BadLoginPopupOperator
from popup_for_account_needed       import   AccountNeededPopupOperator

from sync                           import   PRISTINA_OT_SyncResults


# must do this after ensure exr and load supabase, else get supabase not found when uploader imports sessions that imports supabase
########## SETUP CALLBACK FUNCTION FOR LOG UPLOADER MODULE  ##########
from log_uploader_REUSED            import LogUploader       # to pass the version setting to the LogUploader class
# Callback function (here as a lambda) is registered with, and called by LogUploader, return values are
# used for zipped log filenames in the Supabase 'pristina_logs' bucket -- eg "GOOD_blender_plugin_v0.41_log_2167.zip"
distro = 'blender_plugin'                   # <<<<<<<<<<< CUSTOMIZE HERE,   eg '3ds_plugin' for 3ds max application
version = 'v0.46'  # come back to this todo as now set in 3 places, init and nested panels  and main
LogUploader.callback_for_distro_and_version = lambda: ( distro,  version)
logger.info(f'LogUploader callback now set with:       {distro} {version}')  



########## BLENDER METADATA  'HEADER'  ##########


bl_info = {
    "name":        "Pristina (TM)    (c)2025 Ideal Real Labs",      # todo  change all (tm) marks to (r) when got
    "version":     (0, 0),                 # version etc now set in __init__ but this bl_info must be present else error
    "blender":     (2, 80, 0),
    "category":    "Render",
    "description": "Supercharge your archviz workflow with Pristina AI"
                   " -- turn any image into a professional-grade image",
    "warning":     "*** Installation requires elevated privileges (admin or root) ***",
    "wiki_url":    "https://pristina.ai",      
}


logger.info(f'\n\n\t*******  Pristina Plugin  v{bl_info["version"]} Log. Running file\n\t{__file__}\n\n')





########## REGISTRATION ##########



registration_classes = [
    
    PristinaItem,
    PristinaProperties,
    PRISTINA_OT_AddItem,
    PRISTINA_OT_RemoveItem,
    PRISTINA_OT_MoveItem,
    PRISTINA_OT_SelectObject,
    
    PRISTINA_PT_MainPanel,        
    PRISTINA_PT_GlobalSettings,
    PRISTINA_PT_ObjectSettings,
    PRISTINA_PT_AccountSettings,
        
    PRISTINA_OT_RenderPristina,
  
    PRISTINA_OT_Login,
    PRISTINA_OT_OpenLink,
    PRISTINA_OT_SortByDistance,
    ERROR_OT_popup_message,
    PRISTINA_OT_SyncResults,

    LoadImageFromFileOperator,
    TaskFailedPopupOperator,
    CreditsLowPopupOperator,
    BadLoginPopupOperator,
    AccountNeededPopupOperator,
]




def register():
    logger.info('__init__ register() called')
    
    for cls in registration_classes:
        try:
            logger.info(f'__init__  REGistering  {cls}')
            bpy.utils.register_class(cls)
        except Exception as ex:
            logger.info(f'---------------Registration for {cls} failed: {ex}')
            
          
    bpy.types.Scene.pristina_properties = bpy.props.PointerProperty(type=PristinaProperties)  
    logger.info( f'__init__ register() completed\n')
    
    
    # Defer popup execution to ensure a valid context  (wait and hope?)
    def deferred_popup():
        AccountNeededPopupOperator.one_shot_flag = True
        AccountNeededPopupOperator.do_popup()
        return None  # Stops the timer

    bpy.app.timers.register(deferred_popup, first_interval=0.5)
                
                
    
           




def unregister():

    logger.info('__init__  unregister() called')

    logger.info(f'Setting stop event flag')
    stop_event.set()                        # see fifo q module: event state is regularly checked in the 'wait_for_complete' 
                                            # polling loop -- each 'wait for' has its own thread (ie one per job)
                                            # Each thread will return/complete as as soon as stop event is heard
                                            # and in any case (back stop) the threads are daemon, 
                                            # so will complete when python exits
                                            
    SessionManager.close_all_sessions()     # make sure to close all existing auth sessions with Supabase
                                            # !!! if these logouts are not all done, or one or more fails, 
                                            # then Blender carries on running in the background      
    
    for cls in reversed(registration_classes):
        try:
            logger.info(f'__init__ UNregistering  {cls}')
            bpy.utils.unregister_class(cls)
        except Exception as ex:
            logger.info(f'----------------Unregistration for {cls} failed: {ex}')
            
            
    try:
        del bpy.types.Scene.pristina_properties
    except:
        pass
        
    if Timer_for_FIFO_checking.timer_running:
        Timer_for_FIFO_checking.stop()      # will unregister the associated Blender timer
    logger.info('__init__ unregister() completed')




if __name__ == "__main__":
    register()



'''
Q. CNM
in the blender addons install screen, the icon on the left of the plugin name will trigger a dropdown showing file warning and user (remove). What does the remove do (in detail -- because somehow variables are getting remembered when I expect them to be forgotten)

A. ChatGPT 4o
In Blender, the "Remove" option in the addons install screen performs several actions to fully remove the plugin. Here's a detailed explanation of what happens when you select "Remove" for an addon:

    1.Unregistering Classes and Operators: When you remove an addon, Blender calls the unregister function defined in the addon's Python script. This function typically unregisters any classes, operators, panels, and other components that the addon has registered with Blender.

    2. Removing User Preferences and Settings: Addons can store user preferences and settings in Blender's configuration files. When you remove the addon, these settings are usually deleted from Blender's configuration, ensuring that they do not persist.

    3. Deleting Files: The actual Python script files of the addon are deleted from Blender's addons directory. This prevents the addon from being loaded the next time Blender starts.

    4.Clearing Memory: Any variables, data structures, or objects created by the addon are cleared from Blender's memory. This includes any global variables or persistent data the addon might have been using.

    5. Resetting State: Blender resets its internal state to ensure that any changes made by the addon are reverted. This includes UI changes, added properties, and other modifications.

Despite these steps, sometimes variables or settings might appear to be remembered due to the following reasons:

    Cached Data: Some addons might cache data in temporary files or other locations that are not cleared when the addon is removed. This data can persist between Blender sessions.

    Configuration Files: Addons might write to Blender's configuration files or other persistent storage that is not automatically cleared when the addon is removed.

    Python Modules: If an addon imports other Python modules, those modules might remain in memory even after the addon itself is removed. This can cause some variables or settings to persist.

To ensure that an addon is completely removed and all its variables are forgotten, you can take additional steps:

    Manually Delete Addon Files: After using the "Remove" option, check the addons directory (.../Blender/2.xx/scripts/addons/) to ensure that all files related to the addon are deleted.

    Clear Blender Cache: Clear any cached data that the addon might have created. This might involve manually deleting temporary files or clearing Blender's cache.

    Restart Blender: Restart Blender to ensure that all Python modules and memory are cleared.

Check Configuration Files: Inspect Blender's configuration files (.../Blender/2.xx/config/) to see if the addon has written any persistent settings. Remove any related entries if necessary.



'''