# nested_ui_panels.py

# usage
# from nested_ui_panels   import PRISTINA_PT_MainPanel
# from nested_ui_panels   import PRISTINA_PT_GlobalSettings
# from nested_ui_panels   import PRISTINA_PT_ObjectSettings
# from nested_ui_panels   import PRISTINA_PT_AccountSettings


import bpy
import io
import traceback
import pristina_config  as config       # global settings for credits_required for different features
from   sync             import   PRISTINA_OT_SyncResults
from   logger_module    import   Logger

logger       = Logger.get_logger()      ####################  global, is set up on load 


       

class PRISTINA_PT_MainPanel( bpy.types.Panel):
    
    bl_label       = "Pristina " + "v0.46"  # ***** must be set here and also in __init__.py     ******
                                            # now set in 3 places, init and nested panels  and main
                                            # -- all attempts to get single-setting have failed 
                                            # because Blender also does non-python parsing of sources to get metadata
                                            # https://blender.stackexchange.com/questions/227940/how-to-get-bl-info-dict-into-submodule-if-it-is-possible
                                            
    bl_idname      = "PT_PRISTINA_PT_MainPanel"
    bl_space_type  = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context     = "render"
    bl_category    = "Pristina"         # This is the name of the new tab
    
    
    
    def count_true_properties( object_properties):
        restore_faces_count       = 0
        export_mask_count         = 0
        use_reference_image_count = 0
        ultra_count               = 0

        for obj_name, props in object_properties.items():
            if props.get('restore_faces', False):
                restore_faces_count       += 1
            if props.get('export_mask', False):
                export_mask_count         += 1
            if props.get('use_reference_image', False):
                use_reference_image_count += 1
                
            quality = props.get("quality", 1)   #  "Good")     # value is not set to string, instead it is set to index in list
                                                # 0:Fast, 1: Good,  2:Best,  3:Ultra
            
            ultra = 3                           # index in list for Ultra                
            # logger.critical( f'\n\n\t***** quality={quality} *****\n\n')           # see what is happening       
            if ( quality == ultra ):    
                ultra_count += 1

        return restore_faces_count, export_mask_count, use_reference_image_count, ultra_count
        
        

        

    def calculate_credits_required( context):     
        pristina_props           =   context.scene.pristina_properties
        num_items                =   len( pristina_props.pristina_items)                
        total_credits_required   =   num_items * config.credits_per_item    # 'cover charge', per item regardless of options
                                                                            # total_credits_required is equivalent of 'total_score'
                                                                            # in task processor app, render.py, calculate_scores()        
        upscale_result           = pristina_props.upscale_result    
        
        #if upscale_result:
         #   total_credits_required      += config.credits_to_upscale
         
        
        if upscale_result == 'Fast 2X':                                     # Handle the different upscale options (now dropdown)
            total_credits_required      += config.credits_to_fast_2x

        elif upscale_result == 'Creative 2X':
            total_credits_required      += config.credits_to_creative_2x
            
            
        
        restore_faces_count, export_mask_count, use_reference_image_count, ultra_count\
                                 = PRISTINA_PT_MainPanel.count_true_properties( pristina_props.pristina_items)
        
        total_credits_required          += ( restore_faces_count        * config.credits_to_restore_faces)
        total_credits_required          += ( export_mask_count          * config.credits_to_export_mask)
        total_credits_required          += ( use_reference_image_count  * config.credits_to_use_reference)
        total_credits_required          += ( ultra_count                * config.credits_for_ultra)
        
        # logger.info(f'\n\n\t***** total_credits_required={total_credits_required} *****\n\n')

        return total_credits_required




    
    def draw( self, context):
        
        from task_manager import TaskManager
        
        # 0000000000000000000000000000000000000000000000000000000000000000000000000000000000
        
        
        layout         = self.layout
        scene          = context.scene
        pristina_props = scene.pristina_properties

        num_credits_required  =  PRISTINA_PT_MainPanel.calculate_credits_required(context)
        
        
        # ------------------ do purple disk on rhs of Render Pristina button -------------
        
        split            = layout.split(factor=0.95)        # Create a split layout with two columns
       
        col_left         = split.column()                   # Left column
        row              = col_left.row()
        row.scale_y      = 2.0                              # Adjust height of the button
        
        row.operator( "scene.render_pristina", 
                      icon = "RESTRICT_RENDER_OFF", 
                      text = f"Render Pristina ({num_credits_required})")
       
        col_right        = split.column()                   # Right column for the purple disk
        disk_row         = col_right.row()
        disk_row.scale_y = 2.0                              # Adjust height of the purple disk 
        disk_color       = TaskManager.current_indicator_color
        
        disk_row.template_node_socket( color=disk_color)    # default is Purple color  
        
        
        # ----------------------------- jobs in queue ---------------------------------------
      
        layout.separator()
        box = layout.box()

        box.label( text = f"Jobs in Queue: ({pristina_props.jobs_in_progress})")        
        

    @classmethod                        # usage: eg    PRISTINA_PT_MainPanel.redraw_panel(bpy.context)
    def redraw_panel( cls):
        context = bpy.context
        
        if context.screen is None:      # *** error will occur if redraw() called from non-main thread ***
            mem_file           = io.StringIO()      # Capture the stack trace into a string
            traceback.print_stack( file=mem_file )
            stack_trace_string = mem_file.getvalue()
            # Print the traceback and the error message
            logger.error(  f"\n\n\nredraw_panel() trace:\n{stack_trace_string}\n\n" + 
                            "context.screen is None -- did you call from non-main thread?\n\n\n")
            return                        
        
        for area in context.screen.areas:
            if area.type == 'PROPERTIES':
                area.tag_redraw()   
                     
        bpy.ops.wm.redraw_timer( type='DRAW_WIN_SWAP', iterations=1)  # force immediate redraw of ui  VERY RELIABLE !
        
        
        
        
        
        
         
        # 111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111
        
class PRISTINA_PT_GlobalSettings(bpy.types.Panel):
    
    bl_label        = "Pr Global Settings"
    bl_idname       = "RENDER_PT_global_settings_panel"
    bl_space_type   = 'PROPERTIES'
    bl_region_type  = 'WINDOW'
    bl_context      = "render"
    bl_parent_id    = "PT_PRISTINA_PT_MainPanel"
    bl_description  = "click to hide, click to drop down"
    bl_options = {'DEFAULT_CLOSED'}  # Initially closed by default  <<<<<<<<<<  CLOSED INITIALLY

    def draw(self, context):
        layout         = self.layout
        scene          = context.scene
        pristina_props = context.scene.pristina_properties

        layout.separator()
        layout.prop( pristina_props, "file_output",     text="Render Output")
        layout.prop( pristina_props, "job_name",        text="Job Name")
        layout.separator()
        layout.prop( pristina_props, "globalprompt",    text="Global Scene Prompt")
        layout.prop( pristina_props, "globalnegprompt", text="Global Scene Negative Prompt")

        layout.prop( pristina_props, "upscale_result",  text="Upscale Result")
        layout.separator()
        
        
        # 222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222
        
        
class PRISTINA_PT_ObjectSettings(bpy.types.Panel):
    bl_label        = "Pr Object Settings"
    bl_idname       = "RENDER_PT_object_settings_panel"
    bl_space_type   = 'PROPERTIES'
    bl_region_type  = 'WINDOW'
    bl_context      = "render"
    bl_parent_id    = "PT_PRISTINA_PT_MainPanel"
    bl_options = {'DEFAULT_CLOSED'}  # Initially closed by default  <<<<<<<<<<  CLOSED INITIALLY

    def draw(self, context):
        layout         = self.layout
        scene          = context.scene
        pristina_props = context.scene.pristina_properties
        
        layout.label( icon='SEQ_SPLITVIEW', text="AI Items:")
        
        row = layout.row()
        
        row.template_list( "UI_UL_list", 
                           "pristina_items", 
                           pristina_props, 
                           "pristina_items", 
                           pristina_props, 
                           "active_item_index", 
                           rows=3)

        col = row.column(align=True)
        col.operator( "scene.pristina_add_item",    icon='ADD',             text="")
        col.operator( "pristina.remove_item",       icon='REMOVE',          text="")
        col.operator( "pristina.move_item",         icon='TRIA_UP',         text="").direction = 'UP'
        col.operator( "pristina.move_item",         icon='TRIA_DOWN',       text="").direction = 'DOWN'
        col.operator( "pristina.sort_by_distance",  icon='SEQ_STRIP_META',  text="")
        col.operator( "pristina.select_object",     icon='DOT',             text="")

        if pristina_props.pristina_items:
            
            selected_item = pristina_props.pristina_items[pristina_props.active_item_index]
            
            layout.separator()
            layout.prop_search( selected_item, "object_input", context.scene, "objects", text="Object")
            layout.separator()
            
            box = layout.box()
            
            box.label(text="Object AI Configuration:")
            
            box.prop(selected_item, "prompt",                   text= "Prompt")
            
            box.prop(selected_item, "negative_prompt",          text= "Negative Prompt")
            box.separator()
            box.prop(selected_item, "denoise_strength",         text= "AI Strength")
                                                              
            box.prop(selected_item, "geometry_strength",        text= "Geometry Strength")
            
            box.prop(selected_item, "sd_seed",                  text= "Seed")
            box.separator()
            
            box.separator()
            box.prop(selected_item, "presets",                  text= "Presets")
            
            box.prop(selected_item, "quality",                  text= "Quality")
            
            box.prop(selected_item, "restore_faces",            text= "Restore Human Faces")
            
            box.prop(selected_item, "export_mask",              text= "Export mask")
            
            box.prop(selected_item, "use_reference_image",      text= "Use Reference Image")
            
            if selected_item.use_reference_image:
                box.prop(selected_item, "reference_image_path", text= "Reference Image Path") 
                
                box.prop(selected_item, "reference_strength",   text= "Reference Strength")

        # 3333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333333
        
        
        


class PRISTINA_PT_AccountSettings(bpy.types.Panel):
    bl_label        = "Pr Account"
    bl_idname       = "RENDER_PT_account_panel"
    bl_space_type   = 'PROPERTIES'
    bl_region_type  = 'WINDOW'
    bl_context      = "render"
    bl_parent_id    = "PT_PRISTINA_PT_MainPanel"

    def draw(self, context):
        layout         = self.layout
        scene          = context.scene
        pristina_props = context.scene.pristina_properties
        
        layout.separator()
         # New UI elements for Email and Password
         
        box             = layout.box()
        
        if pristina_props.auth_status:  # Check the authentication status
            # logger.info(f'\n\n\t**** login: got auth status: True *****\n\n')
            row     = box.row()
            row.label( text="Account: Connected", icon='MOD_WAVE')  # Use a check icon or similar
            if pristina_props.user_id:
                row = box.row()
                row.label( text=f"Credits: {pristina_props.credits}")                
        else:
            # Optionally, add an indication for not connected or failed authentication
            row       = box.row()
            row.alert = True  # Set the alert state for this row
            row.label(text="Account: Not Connected", icon='ERROR')
            
        box.prop( pristina_props, "email",          text="Email")
        box.prop( pristina_props, "password_input", text="Password")
        box.operator( "pristina.login",             text="Login")
        
        
        num_credits_required  =  PRISTINA_PT_MainPanel.calculate_credits_required(context)        
        
        box.label( text=f"Credits after render: {pristina_props.credits - num_credits_required}")
        box.operator( "pristina.open_link",   text="Buy More Credits",  icon='URL')        
        

        syncing = PRISTINA_OT_SyncResults.syncing
        button_text = "Syncing..." if syncing else "Sync"
        box.operator(PRISTINA_OT_SyncResults.bl_idname, text=button_text)
        
   
        for area in context.screen.areas:
            if area.type == 'PROPERTIES':
                area.tag_redraw()   
     
        
  
