# ot_render_pristina.py

# from ot_render_pristina   import   PRISTINA_OT_RenderPristina

import bpy
import os
import sys
import json
from task_manager         import TaskManager
from compositor_setup     import CompositorSetup
from datetime             import datetime, timedelta

from object_analyzer      import ObjectAnalyzer
from cryptomatte_renderer import CryptomatteRenderer
from viewport_renderer    import ViewportRenderer
from file_utils           import append_timestamp_to_filename
from file_utils           import PristinaZipper

from logger_module        import Logger
from image_resizer        import ImageResizer

import popup_for_low_credits


logger  = Logger.get_logger()      ####################  global, is set up on load 




class PRISTINA_OT_RenderPristina( bpy.types.Operator):
    
    bl_idname      = "scene.render_pristina"
    bl_label       = "Render Pristina"
    bl_description = "Render Pristina: Render the project and\nsend to Pristina.ai for cloud processing\n(in brackets)=number of credits required"


    def has_enough_credits_to_render( context):
        from nested_ui_panels   import PRISTINA_PT_MainPanel
        
        got_credits = context.scene.pristina_properties.credits
        required    = PRISTINA_PT_MainPanel.calculate_credits_required(context)   

        popup_for_low_credits.check_credits( got_credits - required)   # check if less than threshold (eg 100) after this job
        
        return  got_credits  >=  required
        
        
    def job_has_no_items():
        return False
    
    

    def execute(self, context):        
        from ot_login           import   PRISTINA_OT_Login # technique to break out of circular referencing of imports
        from fifo_queue_module  import   Timer_for_FIFO_checking
        from fifo_queue_module  import   new_job
        
        pristina_props = context.scene.pristina_properties

        num_items      =  len( pristina_props.pristina_items)
        logger.info( f'render pressed, there are {num_items} items')
        
        
        #----------------------------------------------------   blender version and exr checks
        # Blender's installation directory
        python_executable = sys.executable
        blender_install_dir = os.path.dirname(os.path.dirname(python_executable))
        logger.info(f"Blender version in path: {blender_install_dir}")
        # eg  Blender installation directory: C:\Program Files\Blender Foundation\Blender 4.1\4.1\python
        # r"C:\Program Files\Blender Foundation\Blender 4.1\4.1\python\lib\site-packages"   # example path to supabase package

        version_path = os.path.dirname(blender_install_dir)
        exr_expected_at = os.path.join( version_path, 'datafiles', 'studiolights', 'matcap', 'check_normal_pristina.exr')
        logger.debug(f'exr expected at {exr_expected_at}')
        found = os.path.exists( exr_expected_at)
        if found:
            logger.debug(f'exr found in place ok')
        else:
            logger.critical(f'\n\n\n\t\t************* exr not found ***************\n\n\n')
        
        # ----------------------------------------------------------------------------------------------
        
        
        
        if  num_items == 0:                                 # if user presses render button with no items, then just ignore
            return {'FINISHED'}
        

        TaskManager.set_indicator_color('credit_check')     # eg set disk colour to orange---immediate response on button press
        logger.info(f'\n\n\t***** RENDER PRISTINA BUTTON PRESSED *****\n\n')
        
        
        PRISTINA_OT_Login.execute(self, context)


        if (not PRISTINA_OT_RenderPristina.has_enough_credits_to_render( context)):
            self.report({'ERROR'}, "Not enough credits")
            return {'CANCELLED'}
    
        


        Timer_for_FIFO_checking.start( pristina_props)      # timer start method does nothing if already running

        TaskManager.set_indicator_color('in_render')        # eg set disk color to red
        

        output_dir       = bpy.path.abspath( pristina_props.file_output)   

        '''
        # experimental default for when output_dir is still set to ""  
        # -- use same directory as the current .blend we are in
        if output_dir == "":
            output_dir   = os.path.dirname( bpy.context.blend_data.filepath)
        '''

        # todo --- decide if want a default output directory, or always require user to set valid directory

        # experimental default for when job_name is still set to ""  
        # -- use same directory as the current .blend we are in
        job_name        = bpy.path.abspath( pristina_props.job_name)   
        if job_name == "":
            filename_dot_blend = bpy.path.basename( bpy.context.blend_data.filepath)
            filename, _ = os.path.splitext(filename_dot_blend)       # get name of current blend file without the extn
            pristina_props.job_name = filename
            
            logger.critical(f'name set to .blend filename { pristina_props.job_name}')
        


        # this block will not apply if the above block remains (test will fail always)
        if not  os.path.isdir( output_dir):                 # cnm
            bpy.ops.error.popup_message('INVOKE_DEFAULT', message="Please enter valid Render Output folder.")
            TaskManager.set_indicator_color('default')      # set disk color back to purple
            return {'CANCELLED'}                            # todo, in the warning dialog, include a browse dir select

        
        
        
        tmp_dir          = os.path.join( output_dir, "Pristina_tmp")    # Create 'Pristina_tmp' dir in output directory
        os.makedirs( tmp_dir, exist_ok=True)

                                                            # Generate timestamped directory name for "packme" 
                                                            # within the 'Pristina_tmp' folder
        timestamp        = datetime.now().strftime( "%Y%m%d_%H%M%S")
        packme_dir_name  = f"packme_{timestamp}"
        packme_dir       = os.path.join( tmp_dir, packme_dir_name)
        os.makedirs( packme_dir, exist_ok=True)
        
        object_list      = [item.object_input.name   for item   in pristina_props.pristina_items   if item.object_input]
        
        ### Depth map run ##############
        # Example usage
        #depthfileoutput = output_folder
        analyzer         = ObjectAnalyzer( object_list)
        analyzer.analyze_objects()
        compositor_setup = CompositorSetup()
        compositor_setup.setup_nodes( packme_dir)
        
        
        
        # Assuming CryptomatteRenderer and its functionality is defined correctly elsewhere
        renderer         = CryptomatteRenderer( object_list, packme_dir)
        
        # DEPTH CLEANUP After you're done with the nodes and need to clean up
        compositor_setup.cleanup_nodes()
                
        
        
        ### Normal map run
        # matcap_url        = 'https://irlabs.co.uk/ai/check_normal_pristina.exr'
        # matcap_name       = 'check_normal_pristina.exr'
        viewport_renderer = ViewportRenderer(os.path.join(packme_dir, "pristina_normal_map.png"))
        # viewport_renderer.run(matcap_url, matcap_name)
        
        dest = r"C:\Program Files\Blender Foundation\Blender 4.1\4.1\datafiles\studiolights\matcap\check_normal_pristina.exr" # todo remove
        # logger.critical( f'check exr in place  {os.path.exists(dest)}  size={os.path.getsize( dest)}')  # todo remove temp dev only

        viewport_renderer.run()


        ### SAVE References
        for item in pristina_props.pristina_items:
            if item.use_reference_image and item.reference_image_path:
                input_image = bpy.data.images.load(item.reference_image_path)
                output_name = f"{item.object_input.name}_pristina_reference_input.jpg"
                #output_path = pristina_props.file_output
                reference_strength = item.reference_strength
                resizer = ImageResizer(input_image, output_name, packme_dir, reference_strength)
                resizer.resize_and_save()


        file_path = os.path.join(packme_dir, "render_output_properties.json")
        
        object_properties = { item.object_input.name: {

                "geometry_strength":        item.geometry_strength,
                "ai_strength":              item.denoise_strength,
                "color_strength":           0.5,
                "seed":                     item.sd_seed,
                "presets":                  item.presets,
                "quality":                  item.quality,
                "prompt":                   item.prompt,
                "negative_prompt":          item.negative_prompt,
                "export_mask":              item.export_mask, 
                "restore_faces":            item.restore_faces,
                "use_reference":            item.use_reference_image,
                "reference_strength":       item.reference_strength

            } for item in pristina_props.pristina_items if item.object_input
        }
        
        data = {
            "render_output":                os.path.join(packme_dir, pristina_props.file_name + ".png"),
            "Plugin":                       "Pluginclient",
            "User":                         "user@example.com",  # Replace with actual user info
            "Authtoken":                    "auth_token",  # Replace with actual auth token
            "global_prompts": {
                "positive_prompt":          pristina_props.globalprompt,
                "negative_prompt":          pristina_props.globalnegprompt,
                "upscale_result":           pristina_props.upscale_result
            },

            "object_properties": object_properties
        }

        with open(file_path, 'w') as f:
            json.dump(data, f, indent=4)
        
        self.report({'INFO'}, f"Properties written to {file_path}")
        
        zip_file_name                  = "pristina-tmp-output.zip"
        pristina_file_name             = "output.pristina"
        timestamped_zip_file_name      = append_timestamp_to_filename( zip_file_name)
        timestamped_pristina_file_name = append_timestamp_to_filename( pristina_file_name)

        
        zipper = PristinaZipper(    packme_dir, 
                                    tmp_dir, 
                                    timestamped_zip_file_name, 
                                    timestamped_pristina_file_name)
                                    
        zipper.zip_folder()

        output_local_path   = os.path.join(tmp_dir, timestamped_pristina_file_name)
        self.report({'INFO'}, f"Output directory zipped and renamed to .pristina in {tmp_dir}")

        
        # Define the UUID, file path, and properties to be passed to the script
        
        userEmail    = pristina_props.email
        userPassword = pristina_props.password
        
        num_items    = len( pristina_props.pristina_items)                  # CALCULATE WHEN TO TIMEOUT IF NOT COMPLETE OR FAILED
        allow_max_time_per_item = 600                                       # allow 10 mins per item for timeout calculation  (set to 10 seconds for testing)
        timeout_seconds         = num_items * allow_max_time_per_item       # timeout period for job with num_items    

        if pristina_props.upscale_result:                                   # allow extra 10 minutes when upscaling    
            timeout_seconds += 600                                              
            logger.info(f'{num_items} items: timeout after {timeout_seconds//60} mins (after adding 10 mins for upscale)')
        else:
            logger.info(f'{num_items} items: timeout after {timeout_seconds//60} mins')
        
        result       = new_job( output_local_path,   userEmail, userPassword,  context,  timeout_seconds)  #   new_job() >>>> in the  'fifo_queue_module'

        # because each job has its own thread, and it communicates status back to the main thread
        # using the FIFO queue, there is no need to wait for the return value of the new_job() call
        # we know what has happened already.   

        return {'FINISHED'}
        
        
    @staticmethod
    def show_message(message, message_type={'INFO'}):    # todo !!!!!!!!!!! have not been able to get this working yet, want it to show timeout message in upload_task... module
        """
        General function to display a message in Blender's Info area and UI.    
        :param message: The message to display.
        :param message_type: The type of message, can be {'INFO'}, {'WARNING'}, or {'ERROR'}.
        """
        # Ensure the context has a window manager
        context = bpy.context
        if context.window_manager:
            context.window_manager.report(message_type, message)
        else:
            # Fallback to console output if no valid window manager
            logger.info(f"{message_type}: {message} \ncould not be shown in Blender info area")
        