# pip_installs.py


import bpy
import sys
import os
import shutil
import subprocess
import ensurepip
import importlib
import time
from logger_module    import   Logger

logger       = Logger.get_logger()      




packages_to_install = {         # add others here if needed
    'supabase': '2.4.6',
    
}


# Ensure pip is installed
ensurepip.bootstrap()

# Function to install a package using pip
#def install_package( package_name, version):
   #package_spec = f"{package_name}=={version}"
    #subprocess.check_call([sys.executable, "-m", "pip", "install", package_spec])
    
    
# Function to install a package using pip
def install_package( package_name, version):
    try:
        package_spec = f"{package_name}=={version}"
        logger.info(f"Attempting to install package: {package_spec}")
        subprocess.check_call([ sys.executable,     "-m", 
                                                    "pip", 
                                                    "install", 
                                                    #"--ignore-installed",
                                                    package_spec])
        
        
        
        logger.info(f"Successfully installed package: {package_spec}")
    except subprocess.CalledProcessError as ex:
        logger.error(f"Failed to install package {package_name}=={version}. Error: {ex}")
    except Exception as ex:
        logger.exception(f"An unexpected error occurred while installing package {package_name}=={version}. Error: {ex}")    
    

# Function to uninstall a package using pip
def uninstall_package( package_name):
    subprocess.check_call([sys.executable, "-m", "pip", "uninstall", "-y", package_name])

# Function to get the installed version of a package
def get_installed_version( package_name):
    try:
        package = importlib.import_module( package_name)
        logger.warning(f'found {package_name} version {package.__version__} already installed')
        return package.__version__
    except ImportError:
        return '[not found]'   # not found is not an error here, expect not to find packages before install


            
def copy_exr_file_to_matcap():                     # this is called in __init__.py   in register()
    
    try:
        plugin_dir  = os.path.dirname(__file__)
        source      = os.path.join( plugin_dir, "check_normal_pristina.exr")
         
        if not os.path.exists(source):
            raise RuntimeError(f"Could not find .exr {source}")

        install_dir = bpy.utils.resource_path('LOCAL')
        matcap_dir  = os.path.join( install_dir, "datafiles", "studiolights", "matcap")

        if not os.path.exists( matcap_dir):  os.makedirs( matcap_dir)

        dest = os.path.join( matcap_dir, os.path.basename(source))


        # Check if destination file exists and has the same size and modification time
        if os.path.exists( dest):
            source_stat = os.stat( source)
            dest_stat   = os.stat( dest)
            if (source_stat.st_size  == dest_stat.st_size    and
                source_stat.st_mtime == dest_stat.st_mtime):                    
                logger.info(f"   .exr file is already in place and is up-to-date: \n{dest}")
                return


        shutil.copy2(source, dest)  # shutil.copy2 preserves the metadata
         
        # Install and register the custom studio light
        #bpy.utils.studiolight_install( filepath=dest, make_paths_relative=True, name="check_normal_pristina")
        #logger.info("Studio light installed successfully.")

        # Set the installed studio light
        #bpy.context.space_data.shading.studio_light = "check_normal_pristina"
        #logger.info(f"Studio light set to 'check_normal_pristina'")
        
        # List available studio lights
        #available_lights = bpy.types.SpaceView3D.bl_rna.properties['studio_light'].enum_items.keys()
        #logger.info( "Available Studio Lights:", list(available_lights))
                    
                      
        logger.info(f"\n\nCopied \n{source} to \n{dest}\n\n")
        return     # operations were successful
        
    except PermissionError as ex:
        logger.warning(f'\n\n\t*** Pristina install requires elevated privileges (admin or root) to avoid permission errors ***\n\n')
        logger.info(f'\n\n\t***** Now quitting Blender *****\n\n')

    except Exception as ex:
        logger.exception(f'copy exr unexpected exception: {ex}')     
        
    finally:      
        pass        #  comment out next line to see any exception/error messages that would disappear on quick exit
        #os._exit(99)                # should shut Blender immediately 
            
            
def ensure_dependencies_installed():
    try:
        # Install packages as per the dictionary
        for package, version in packages_to_install.items():
            installed_version = get_installed_version( package)
            
            if installed_version != version:
                if installed_version is not None:
                    logger.info( f"Uninstalling {package} version {installed_version}")
                    uninstall_package( package)
                logger.info( f"Installing {package} version {version}")
                install_package( package, version)
                time.sleep( 5 )                         # Add a delay to allow the installation to complete
                importlib.invalidate_caches()           # Invalidate the import caches

        
        for package in packages_to_install.keys():      # Verify installation
            try:
                module = importlib.import_module( package)
                logger.info( f"{package} version: {module.__version__} installed ok")
            except ImportError as ex:
                logger.error( f"{package} installation failed  {ex}.")    
                raise
                
        logger.info( f'\n\n\t***** Dependencies verified ok *****\n\n')
                
    except Exception as ex:
        logger.critical( f'\n\n\n**** COULD NOT ENSURE DEPENDENCIES INSTALLED *****\n\nException {ex}\n\n\n')
        raise
            


# ensure_dependencies_installed()         # >>>>>>>>>> this code executes immediately on load >>>>>>>>>>>>>>>>>    NOT CURRENTLY:
            
            
            
############### CURRENTLY USING THIS MUCH SIMPLIFIED CODE BELOW, THAT 'WORKS'      
############### TODO, COME BACK TO THIS LATER AND IMPROVE USING PREVIOUSLY DEVELOPED CODE ABOVE THAT IS VERSION AWARE ETC ###############      
############### MAJOR UPGRADE HERE IS TO USE THE "--ignore-installed" OPTION IN PIP INSTALL  in install_package() method --->
############### TO PREVENT 'requirement already satisfied' from global python/version site-packages,  
############### in which case, packages such as supabase are not installed into Blender, and give import not found errors: this fixes that problem

logger.info(f'installing supabase with pip')
install_package( 'supabase', '2.4.6')
time.sleep(5)
logger.info(f"found installed version = {get_installed_version( 'supabase')}")
try:
    import supabase
    logger.info(f'import supabase --- *** tested ok ***')
except Exception as ex:
    logger.error(f'error trying to import supabase: {ex}')
    time.sleep( 5)   # allow log to be written
    # os._exit(1)     # quit
    
    
### todo  --- in user instructions -- warn them that when you click the checkbox to install, the tick will not show until the full install is complete, just wait while this happens
### sometimes, because it can take a while to do the pip installs, Blender may temporarily say 'not responding' as if it has hung up, then when the installs finish, it recovers ok

    


