# popup_for_account_needed.py

# based on EULA_check_agree.py

'''
This script may look *unbelievably complex* to just launch a popup -- if you find a simpler way, please let cnm know!
The problem with the usual ways of doing this is that when Blender is starting up and when it is doing add-on installs
it goes into a 'restricted context' mode. The normal ways of doing popups then fail -- with restriction/context/view_layer
errors. The script below manages to set up a valid context regardless, and it works.
'''

import bpy  
import sys
import os
import webbrowser
from logger_module import Logger

logger = Logger.get_logger()  ####################  global, is set up on load 

# =================================================================

# Get visibility for the package and modules --- append sys path
plugin_dir = os.path.dirname(__file__)
if plugin_dir not in sys.path:
    sys.path.append(plugin_dir)

from logger_module import Logger  # Setup logger
logger = Logger.get_logger()




class AccountNeededPopupOperator(bpy.types.Operator):
    """Operator to display the Account Needed popup."""
    bl_idname = "wm.account_needed_popup"
    bl_label = "Notice: Account Needed"

    one_shot_flag = False  # Class attribute to manage the popup trigger state

    @classmethod
    def do_popup(cls):
        """Class method to initiate the popup."""
        logger.info(f'***** Popup flag = {cls.one_shot_flag} *****\n\n')

        if cls.one_shot_flag:
            bpy.app.timers.register(cls.trigger_dialog, first_interval=0.1)

    @classmethod
    def trigger_dialog(cls):
        """Class method to trigger the modal dialog in any suitable context."""
        if cls.one_shot_flag:
            cls.one_shot_flag = False  # Reset the flag immediately to avoid repeated triggering
            logger.info("Trying to invoke Account Needed dialog...")

            # Find a suitable context for the popup
            suitable_area = None
            valid_area_types = ['PREFERENCES', 'VIEW_3D', 'TEXT_EDITOR', 'NODE_EDITOR', 'PROPERTIES']

            for window in bpy.context.window_manager.windows:
                for area in window.screen.areas:
                    if area.type in valid_area_types:
                        suitable_area = (window, area)
                        break
                if suitable_area:
                    break

            if suitable_area:
                window, area = suitable_area
                logger.info(f"Using context: Window={window}, Area={area.type}")
                with bpy.context.temp_override(window=window, area=area):
                    bpy.ops.wm.account_needed_popup('INVOKE_DEFAULT')
                    
            else:
                logger.info("No suitable context found, cannot trigger dialog.")
                
                
            # ----------------------

              
            # Check if the Preferences window is open and close it
            for window in bpy.context.window_manager.windows:
                for area in window.screen.areas:
                    if area.type == 'PREFERENCES':
                        with bpy.context.temp_override(window=window, area=area):
                            bpy.ops.wm.window_close()  # Close the Preferences window
                            print("Preferences window closed.")
                            break
                                        
            # ----------------------

                

        return None  # Stops the timer from repeating

    def draw(self, context):
        """Defines the UI of the popup."""
        layout = self.layout

        # Create a red alert box
        box = layout.box()
        box.alert = True  # Makes the box appear red
        row = box.row()
        row.label(text="Important!", icon='ERROR')  # Adds a warning/error icon

        # Add your text inside the box
        box.label(text="You need an account to use Pristina.")
        box.label(text="Click 'OK' to go to the web sign-up page.")
        box.label(text="")
        box.label(text="(If you already have a username and password, ")
        box.label(text="just click 'Cancel' to continue.)")

    def invoke(self, context, event):
        """Handles invocation of the popup dialog."""
        return context.window_manager.invoke_props_dialog(self)

    def execute(self, context):
        """Handles execution of the popup logic."""
        url = "https://www.pristina.ai/register/"
        try:
            webbrowser.open(url)
            logger.info('Browser launched for user account registration')
        except Exception as e:
            logger.error(f'Failed to launch browser: {e}')

        return {'FINISHED'}

'''
# -------------------------------------------------------------- REGISTRATION

registration_classes = [
    AccountNeededPopupOperator,
]


def register():
    """Registers the operator and launches the popup."""
    logger.info('register() called')

    for cls in registration_classes:
        try:
            bpy.utils.unregister_class(cls)
            logger.info(f'Unregistered {cls}')
        except Exception as ex:
            pass

        try:
            bpy.utils.register_class(cls)
            logger.info(f'Registered {cls}')
        except Exception as ex:
            logger.info(f'---------------Registration for {cls} failed: {ex}')
            pass  # In case comment out/remove logger

    logger.info('Launching Account Needed popup')
    AccountNeededPopupOperator.one_shot_flag = True  # Set the flag on the class
    AccountNeededPopupOperator.do_popup()  # Trigger the popup using the class method


def unregister():
    """Unregisters the operator."""
    logger.info('Account Needed popup unregister() called')
    for cls in reversed(registration_classes):
        try:
            bpy.utils.unregister_class(cls)
            logger.info(f'Unregistered {cls}')
        except Exception as ex:
            pass


 '''