# retrex_decorator___retry_with_exception_catching.py

# usage
# from Supabase.retrex_decorator___retry_with_exception_catching    import * 


import os
import sys
import httpx
import httpcore
from   datetime                    import datetime
import postgrest
import postgrest.exceptions        # no need to install, already in standard supabase python package
#import requests                    # for diy POSTs, as problems with Supabase Python library
import time
import traceback
import subprocess
import random
import json
import zipfile


from logger_module import Logger

logger   = Logger.get_logger()

DEFAULT_NUM_RETRIES = 0


def get_formatted_traceback( ex):
    """
    Extracts and returns a list of dictionaries, each with details of each step in the traceback
    Returns: A list of dictionaries with filename, line number, function name, and line text for each traceback step
    
    NB top of the stack (in this situation) will be in this module, in the wrapper method, giving eg dictionary:
    {'filename': ' ... \\project_pristina\\Supabase\\retrex_decorator___retry_with_exception_catching.py', 
     'lineno': 87, 
     'funcname': 'wrapper', 
     'line': 'result      = target_fn(  *args,   **kwargs) # call target function with positional and keyword args'}
    
    Therefore skip this in the returned stack dictionary list
    """
    tb             = ex.__traceback__
    full_stack     = traceback.extract_tb(tb)
    details_list   = []
    
    for frame in full_stack:
        filename, lineno, funcname, line = frame
        
        # trim down the filename to essentials -- parent/file.extn
        file_extn  = os.path.basename( filename)
        dir_path   = os.path.dirname(  filename)
        parent     = os.path.basename( dir_path)
        filename   = os.path.join( parent, file_extn)         
        
        details    = {
            "filename": filename,
            "lineno":   lineno,
            "funcname": funcname,
            "line":     line
        }
        details_list.append(details)
        
    formatted = "\n"
    for ix, d in enumerate ( details_list[1:]):   # skip the top of list, as note above
        formatted += f'\n\ttraceback{ix}       {filename}    >>>>> in {funcname}()   line {lineno}\n\tline {lineno}: {line}\n\n'
    return formatted






# custom decorator to do retries and catch exceptions, just apply   @retrex()   above a function definition to wrap it
# !!!!!!!!!!!!!!!  Essential to remember to include the brackets so:  @retrex(), otherwise
# you will get failures with error message saying for example: 'wrong number of arguments, expected 1, got 4'  
# also caller functions must themselves have return values in the custom dictionary format seen in many functions,
# where if there has been no exception, the return might be    
#      return { 'data': data,     'error': False,  etc... }
# if there is an error, then retrex will pick it up, give it an explanatory message and give a dictionary return
# with 'error': True,  'message': helpful_hints,  ...}


def retrex(                                 # to override defaults, say eg    @retrex( retries= 10, wait=0.5)
            retries             = DEFAULT_NUM_RETRIES,        
                                            # default how many times to retry calling the target function before giving up
            wait0               = 0.1,      # default initial wait time, constant for linear mode, grows in exponential mode
            mode                = 'exponential',   # default mode is 'exponential' (backoff applies) or 'linear' (same waits)
            backoff_factor      = 2,        # default for exponential, successively doubles wait time if calls are failing
            show_traceback      = True,
            jitter              = True):    # make retries irregular to help solve the 'thundering herd problem' see link:
            
            # https://www.linkedin.com/pulse/tackling-thundering-herd-problem-distributed-systems-ujjwal-gupta
         
         
    def decorator_retrex( target_fn ):      # target_fn -- function being wrapped here for retry attempts, and error checking
        
        fn_name = target_fn.__name__        # get the name of target function as text to use in error messages
        
        def wrapper( *args, **kwargs ):
        
            # args   --  positional argument supplied as list  [arg1, arg2, arg3, ... ]  can be any number of args
            # kwargs --  keyword arguments supplied as dictionary eg {'val1'=5, 'val2'=7}
            
            #------------------------------------------------------------------             
                
            # wrapper for any function that returns a dictionary with key for 'error' whose value may be True or False
            # the target function is called with positional and keyword arguments as supplied
            # if no error then result is returned
            # else 'retries=' retries are made, either in linear (constant) wait time, or with exponential backoff
            
            # THIS WRAPPER ENCAPSULATES NEARLY ALL THE EXCEPTION CHECKING IN THE APP
            # NO NEED FOR EACH HTTP / SUPABASE CALL TO HAVE ITS OWN ERROR CHECKING, JUST @retrex it, AND JOB IS DONE (HERE)
            
            attempt         = 0
            result          = None
            wait_time       = wait0

            while  attempt <= retries:    

                try:            
                    logger.debug( f'[pr] fn={target_fn.__name__}  retries={retries} attempt={attempt}')
                    
                    
                    # ................... below is what is being wrapped .........................
                    
                    result      = target_fn(  *args,   **kwargs)        # call target func with positional and keyword args
                    
                    # ................... above is what is being wrapped .........................
                    
                    
                    
                    success     = isinstance( result, dict)    and     not result[ 'error']
                    
                    #*********************
                    if success:
                        logger.debug( f'[pr] fn={target_fn.__name__}  SUCCESS on attempt={attempt}')
                        return result                                   # return happy, value is pristina return dictionary
                    #*********************

                    else:
                        logger.warning('')                              # blank line in log
                        logger.warning(f'[pr] Retry {attempt} ')        # only warning level till run out of retries, then error
                        
                                

                    if attempt < retries:                               # Only sleep if there are retries left
                    
                        if jitter:                            
                            jitter_amount = wait_time * random.random() # Apply jitter by randomizing the
                            jittered_wait = wait_time + jitter_amount   # additional wait time up to 100% more
                            
                        else:
                            jittered_wait = wait_time

                        logger.debug(f'[pr] Sleeping for {jittered_wait:.2f} seconds before retrying...')
                        time.sleep( jittered_wait)

                        if mode == 'exponential':
                            wait_time *= backoff_factor                 # default is to double wait on each retry

                    attempt += 1
                    
                    
                    
                #  *****   catch all exceptions  and say what type they are (wherever possible)  *****
                
                except FileNotFoundError as ex:
                    msg = f'[pr] "The file was not found at the specified path.": {ex}'
                    logger.error( msg)
                    return  {'result':   None,   
                             'error':    True,  
                             'message':  msg}


                    
                except PermissionError as ex:
                    tb = ''
                    if show_traceback:
                        tb  = get_formatted_traceback( ex)   # list of frame summary dictionaries from traceback stack
                    msg = f'[pr] Permission denied: Unable to read/write file: {ex}' + tb
                    logger.error( msg)
                    logger.error( msg)
                    return  {'result':   None,   
                             'error':    True,  
                             'message':  msg}
                    
                    
                    
                except ValueError as ex:
                    tb = ''
                    if show_traceback:
                        tb  = get_formatted_traceback( ex)   # list of frame summary dictionaries from traceback stack
                    msg = f'[pr] eg Invalid URL format / JSON decode error / etc: {ex}' + tb
                    logger.error( msg)
                    return  {'result':   None,   
                             'error':    True,  
                             'message':  msg}
                                                
                                                
                                               
                except RuntimeError as ex:                              # errors that don't fall in any other categories 
                    tb = ''
                    if show_traceback:
                        tb  = get_formatted_traceback( ex)   # list of frame summary dictionaries from traceback stack
                    msg = f'[pr] Exception: {ex}' + tb                  # raised in pr code to give explanation text to error
                    logger.error( msg)
                    return  {'result':   None,   
                             'error':    True,  
                             'message':  msg}
                             
                             
                             
                except zipfile.BadZipFile as ex:                           
                    tb = ''
                    if show_traceback:
                        tb  = get_formatted_traceback( ex)   # list of frame summary dictionaries from traceback stack
                    msg = f'[pr] Bad zip file: {ex}' + tb                 
                    logger.error( msg)
                    return  {'result':   None,   
                             'error':    True,  
                             'message':  msg}
                    
                # removed block that depended on 'requests'


          
                except httpcore.ConnectError as ex:                     # httpcore
                    logger.debug(f'Traceback for bad internet--\n{str(ex)}')
                    msg =  f'\n\n\n\t\t  ***** [pr] BAD OR UNAVAILABLE INTERNET CONNECTION  (httpcore) *****\n\n\n'
                    logger.warning( msg)
                    return  {'result':  None,   
                             'error':   True,  
                             'message': msg}
                                 
                                
                             
                except httpx.ConnectError as ex:                        # httpx
                    tb = ''
                    if show_traceback:
                        tb  = get_formatted_traceback( ex)   # list of frame summary dictionaries from traceback stack
                    logger.debug(f'Traceback for bad internet--\n{str(ex)} {tb}')
                    msg =  f'\n\n\n\t\t  ***** [pr] BAD OR UNAVAILABLE INTERNET CONNECTION  (httpx) *****\n\n\n'
                    logger.error( msg)
                    return  {'result':  None,   
                             'error':   True,  
                             'message': msg}
                                        
                  
                except httpx.RequestError as ex:
                    # simpler, use this if fancy {ex.request.url!r} is problematic 
                    # (tries to solve problem of backslashes getting taken as escaping characters, etc)
                    # msg =  f'[pr] HTTP Request error:  {ex}'   
                    
                    msg = f"[pr] An HTTP error occurred while requesting {ex.request.url!r}: {str(ex)}"
                    #logger.warning( msg)
                    logger.error( msg)
                    return  {'result':  None,   
                             'error':   True,  
                             'message': msg}
                                                       
                    
                except httpx.HTTPStatusError as ex:
                    logger.error( f"HTTP error occurred: code {e.response.status_code}" )
                
                    if ex.response.status_code == 400:
                        msg = f'[pr] 400 Bad Request Error: {str(ex)}'
                        logger.error( msg)
                        return  {'result':   None,   
                                 'error':    True,  
                                 'message':  msg}
                        
                    elif ex.response.status_code == 404:
                        msg = f'[pr] 404 Not Found Error: {str(ex)}'
                        logger.error( msg)
                        return  {'result':   None,   
                                 'error':    True,  
                                 'message':  msg}
                        
                    else:
                        msg = f'[pr] HTTP Error Occurred: {str(ex)}'    # catch all for any other http exceptions
                        logger.error( msg)
                        return  {'result':   None,   
                                 'error':    True,  
                                 'message':  msg}
                                 
           
                    
                except postgrest.exceptions.APIError as ex:
                    tb = ''
                    if show_traceback:
                        tb  = get_formatted_traceback( ex)   # list of frame summary dictionaries from traceback stack
                    # also see https://supabase.com/docs/guides/storage/debugging/error-codes
                    code    = f"   [sb] API Error {   ex.code}"
                    message = f":\n\n\t\t  message:  {ex.message}"
                    details = f",\n\t\t  details:  {  ex.details}"
                    hint    = f",\n\t\t  hint:     {  ex.hint}\n"
                    msg = code + message + details + hint   + tb                                             
                    logger.error( msg)
                    return { 'result':   result,
                             'error':    True,
                             'message':  msg}    

                
                
                except subprocess.CalledProcessError as ex:
                    tb = ''
                    if show_traceback:
                        tb  = get_formatted_traceback( ex)   # list of frame summary dictionaries from traceback stack
                    msg =  f"[pr] Subprocess failed with error: { str(ex)}\n"
                    msg += f"Standard output: {ex.stdout}\n"
                    msg += f"Standard error:  {ex.stderr}\n"
                    msg += tb
                    logger.error( msg)
                    return  {'result':   None,   
                             'error':    True,  
                             'message':  msg}
                
                     
                                
                except Exception as ex:                                 # general catch all for ANY other exceptions
                    tb = ''
                    if show_traceback:
                        tb  = get_formatted_traceback( ex)   # list of frame summary dictionaries from traceback stack
                    msg = f'[pr] {fn_name}\n\t\t\t\tRetry {attempt} Unexpected Exception {str(ex)}'
                    msg += tb
                    logger.exception(f'\n\n\n\n{msg} -- will try for json interpretation, and show if available\n\n\n\n\n\n')  
                    logger.debug(f'\n\n\t  !!!!!!!!!!!!!!!! check the traceback just above this !!!!!!!!!!!!!! \n\n\n')
                        
                    #------------------------------  Guess error message may be JSON and try to convert to dictionary
                    json_string = str(ex)
                    try:
                        # Replace single quotes with double quotes for valid JSON format
                        json_string = json_string.replace("'", '"')
                        data        = json.loads(json_string)
                        msg         += f'\n\nDecoded json exception = {data}\n'                    
                            
                    except json.JSONDecodeError as e:
                        pass
                        # logger.debug("Failed to decode presumed JSON:", str(e))
                        
                    except Exception as e:
                        logger.debug("Non-critical: An error occurred during trial json decoding:", str(e))
                        
                    # -----------------------------    
                        
                        
                    logger.error( msg)
                    return  {'result':   None,   
                             'error':    True,  
                             'message':  msg}
                             
                             
                
                
            # if get to here, then all retries have failed, 
            # the loop was exited, we need to return the last result obtained
            # (which will have 'error' set True, and have hopefully, a useful 'message' string)
            
            logger.error(f'[pr] fn={target_fn.__name__}   Gave up after {retries} retries')
            
            return  result      # which is already in standard format: a dictionary with keys 'error' and 'message'

                

            #------------------------------------------------------------------  

        return wrapper          # structuring in this way allows Python to understand we are building a decorator

    return decorator_retrex     #                              (there is no need for extra Python keywords)
                
                
'''
see
https://stackoverflow.com/questions/61463224/when-to-use-raise-for-status-vs-status-code-testing
for discussion about http status codes and when to use raise for status etc

'''





'''
except requests.exceptions.RequestException as ex:                 
    msg = f'[pr] Network-related error: {ex}'
    logger.error( msg)
    return  {'result':  None,   
             'error':   True,  
             'message': msg}
             
    
except requests.exceptions.Timeout as ex:                 
    msg = f'[pr] Request timed out: {ex}'
    logger.error( msg)
    return  {'result':  None,   
             'error':   True,  
             'message': msg}
             
                
except requests.exceptions.HTTPError as ex:             # catches exceptions from raise_for_status() calls
    msg = f'[pr] HTTP error: {ex}'
    logger.error( msg)
    return  {'result':  None,   
             'error':   True,  
             'message': msg}
             
                
'''        
         