# sync.py

'''
It can happen that a plugin job or jobs is interrupted, the plugin is shut down, 
but Supabase job status then goes to COMPLETE and the job results are not then downloaded

This sync module addresses that by allowing a check when the 'sync' button to see if there
are any COMPLETEs for this user, and if so, catch up on the downloads

The user_jobname which is private to the user is persisted on the user's machine in the form
of the naming of the results folder, empty, waiting for job to finish and being location for 
results downloads.

'''

try:
    import bpy
except ImportError:
    pass                # to allow testing particular methods outside of Blender
    
import os
import queue
import time
import threading
import json 
from   sessions                                          import SessionManager
from   retrex_decorator___retry_with_exception_catching  import retrex
from   file_utils          import   download_and_unzip,   set_status_to_fetched  
from   logger_module       import   Logger


logger     = Logger.get_logger()
                  
TESTMODE   = False

sync_queue = queue.Queue()                 # Define a thread-safe FIFO queue



class PRISTINA_OT_SyncResults(bpy.types.Operator):
    bl_idname      = "pristina.sync_results"
    bl_label       = "Sync"
    bl_description = (
                    "Sync: when submitting a job, once the indicator turns yellow, "
                    "you may safely quit Blender without waiting for the results. "
                    "Use this button to start downloading (in the background) any previous images"
    )

    syncing    = False
    
    

    def execute(self, context):                                     # EXECUTE WHEN SYNC BUTTON IS PRESSED            

        if PRISTINA_OT_SyncResults.syncing:
            return {'FINISHED'}                                     # Ignore multiple presses while syncing        

        logger.info("Sync button pressed: setting button text to 'Syncing...'")
        PRISTINA_OT_SyncResults.syncing    = True                   # Set syncing state and update UI to show "Syncing..."
        self.force_ui_update()                                      # initial setting called from main ui thread, 
                                                                    # (revert is done from thread via queue)

                                                                    # Start the sync operation in a new background thread
                                                                    
        Timer_for_sync_FIFO_checking.start()                        # start the timer loop that check for messages in the sync_queue
                                                                
        thread = threading.Thread(  target = PRISTINA_OT_SyncResults.sync_operation,   
                                    args   = ( context, ))
        thread.daemon = True                                        # in case thread running when quitting Blender, ensures stopped
        thread.start()
            
        return {'FINISHED'}                                         # NB this will return and the 'self' vanish as soon as 
                                                                    # the thread has been spun off. So self cannot be used in other code
                                                                    # therefore set methods as @classmethod  (no need for instance)
                                
      
        
    # ************************ THIS  METHOD RUNS IN ITS OWN DEDICATED THREAD (as background op) ********************

    @classmethod    
    def sync_operation( cls, context):        
        logger.info("sync_operation: Sync operation started in own thread.")
        fetch_any_unfetched_results( context) 
        logger.info("sync_operation: Sync operation complete. Putting SYNC_DONE message on the queue")   
        sync_queue.put( 'SYNC_DONE')                                # send completion signal to the main thread (gui)
        
    # ************************ ABOVE METHOD RUNS IN ITS OWN DEDICATED THREAD (as background op) ********************
        
        

    @classmethod
    def revert_to_idle( cls):
        logger.info("Sync complete, resetting button text to 'Sync'")
        cls.syncing    = False                                      # change button text from 'syncing' back to 'sync'
        cls.force_ui_update()
    
    
    @staticmethod
    def force_ui_update():        
        logger.info("force_ui_update: Forcing UI redraw.")
        for area in bpy.context.window.screen.areas:                # Force redraw of properties panel to update the button text
            if area.type == 'PROPERTIES':
                area.tag_redraw()
                
                
                




class Timer_for_sync_FIFO_checking:  # uses class methods, singleton like, no need to instantiate object to call methods

    interval            = 0.2
    timer_running       = False


    @classmethod
    def start( cls):
        if cls.timer_running:  # may well be running already, don't try and start again
            logger.debug("Timer already running, not starting again")
            return       
        cls.timer_running       = True
        bpy.app.timers.register( cls.check_sync_fifo_now)
        logger.info( f'\n\n\t*****  11111111111111  SYNC FIFO check timer started *****\n\n')



    @classmethod
    def check_sync_fifo_now( cls):
        try:            
            if not sync_queue.empty():     
                try:            
                    status = sync_queue.get_nowait()             
                    logger.info(f'status={status}')
                    if status == 'SYNC_DONE':
                        PRISTINA_OT_SyncResults.revert_to_idle()  
                        cls.timer_running = False  # Ensure timer_running is reset here
                        logger.info( f'\n\n\t***** 00000000000000  SYNC FIFO check timer stopped *****\n\n\n\n')
                        
                        return None             # **This is all that is needed** to completely unregister the timer 
                                                # ie returning 'None' as time interval to next run
                                                # https://docs.blender.org/api/current/bpy.app.timers.html
                                                # any further attempt to unregister this timer will raise an exception                        
                except queue.Empty:
                    pass 
            
        except Exception as ex:
            logger.exception(f"\n\n\t***** Exception in check_sync_fifo_queue: {ex} *****\n\n")
        # logger.info(f'returning interval = {cls.interval}')
        return cls.interval                     # Return interval (in secs) before being called again






@retrex()
def  fetch_any_unfetched_results( context):    
    results_folder  = get_results_folder( context)    
    logger.info(f'Starting FETCH thread, which downloads any results not yet fetched to {results_folder}')
    
    
    # standard approach here: get method running in a new thread, 
    # then start a Supabase session dedicated to that thread
    # and logout as soon as work is done, allowing thread to complete
    
    pristina_props = context.scene.pristina_properties                      # make sure SessionManager is set up
    userEmail      = pristina_props.email
    userPassword   = pristina_props.password    
    if SessionManager.email is None:
        SessionManager.setup( userEmail, userPassword)
    session_name   = SessionManager.get_thread_id_with_name( 'Fetch' )      # get a dedicated supabase session for doing fetching
    
    max_checks     = 100                        # limit just in case loop fails to exit itself,  
                                                # allows for 'max_checks' unfetched jobs to be fetched
    checks_done    =   0
    
    while checks_done < max_checks:
        
        checks_done += 1
        
        # ---------------------------------- GET NEXT COMPLETED RESULTS --------------------------------------------
        
        logger.info( f'\n\n\t***** SEEKING NEXT UNFETCHED RESULT *****\n\n')


        #py# py_fetch_the_unfetched
        

        data = get_oldest_complete( session_name, testmode=TESTMODE)  
        # logger.critical(f'returned ok from get oldest with data={data}')
        
        task_id      = data['task_id']
        if  task_id  is None:                       # no more results to get  >>>>>>>>>>>>>  NORMAL EXIT HERE
            logger.info( data['message'])
            logger.info(f'logging out of fetch session')
            SessionManager.logout( session_name)    # close the 'Fetch' session for this sync operation
            sync_queue.put( 'SYNC_DONE')    
            return {'error': False}                 # no-error, all done, no more COMPLETE jobs 
                                                    # that need downloading and setting to FETCHED
     
       
        # ----------------------FOUND AN UNFETCHED RESULTS SET, DOWNLOAD AND UNZIP (in the normal way) -----------

        storage_path = data['zip_path']
        # logger.info(f'about to download and unzip task {task_id} from {storage_path}')
        
        download_and_unzip( storage_path,  results_folder,  session_name)
              
        #1# 
        
        continue                                    # (for readability only)


    # ----------------------------------------------  if got down to here, then we have timed out, regard as error
    msg = 'timed out after {max_checks} loops'
    logger.info(f'logging out of fetch session')
    SessionManager.logout( session_name)            # ensure even on timeout, session is not left running in the background
    logger.error( msg)
    return { 'error': True,
             'message': msg}
             
             

def get_results_folder( context):
    results_folder = context.scene.pristina_properties.file_output
    if results_folder.startswith("//"):        
        abspath = bpy.path.abspath( results_folder)
    else:        
        abspath = results_folder                        # It's already an absolute path  eg "C:\\Users\\cex\\Documents"        
    return os.path.normpath( abspath )                  # remove any .. relative addressing  eg aa\xx\..\yy --> aa\yy
             



@retrex( retries=1)        
def get_oldest_complete( session_name, testmode=TESTMODE):   
    
    
    
    # ****** THIS PERMISSION MUST BE IN PLACE *********
    
    '''
    -- POLICY user allow UPDATE own row
    -- used only for setting status to FETCHED after download of results

    -- Drop the existing policy if it exists
    DROP POLICY IF EXISTS "USER Allow UPDATE own row" ON "public"."task_queue";

    -- Recreate the policy to allow users to update only their own rows
    CREATE POLICY "USER Allow UPDATE own row"
    ON "public"."task_queue"
    FOR UPDATE
    TO authenticated
    USING (auth.uid() = user_id)
    WITH CHECK (auth.uid() = user_id);
    '''
    
   
    # THIS FUNCTION MUST BE AVAILABLE IN SUPABASE
    # copy this sql code into the sql editor in supabase, 
    # and (bottom right green button) 'Run selected' or kb: ctrl/rtn
    

    '''
    -- FUNCTION  get_oldest_complete

 
    -- Drop the existing function if it exists
    DROP FUNCTION IF EXISTS public.get_oldest_complete( boolean, text);

    -- Create the simplified function
    CREATE OR REPLACE FUNCTION public.get_oldest_complete(
        testmode BOOLEAN,
        testing_account_email TEXT
    )
    RETURNS TABLE (task_id bigint, render_value TEXT) 
    SET search_path = ''
    AS $$
    BEGIN
        -- Retrieve the oldest complete job without updating it
        RETURN QUERY
        SELECT id, render::TEXT
        FROM public.task_queue
        WHERE status = 'COMPLETE'
          AND (
              CASE
                  WHEN testmode THEN user_email = testing_account_email
                  ELSE               user_email <> testing_account_email
              END
          )
        ORDER BY id ASC
        LIMIT 1;
    END;
    $$ LANGUAGE plpgsql SECURITY INVOKER;

    '''
    
    sb_client    = SessionManager.get_client( session_name)         # get fresh client every time, in case it changed on expiry
    # logger.critical('============= got to here ========================')
    params = {
        "testmode":               testmode,                         # or False for production  ----- CHOOSES TESTMODE OR NOT
        "testing_account_email": "test@test.com"
    }
    result       = sb_client.rpc( "get_oldest_complete", params).execute()  
    # logger.critical(f'result=\n\n{ result} AND  len result.data= {len(result.data)}\n\n')
    
    ########################################################
    # result object looks like this
    #
    #         data=[{'task_id':     702, 
    #                'render_value': '{
    #                     "image": "702/702filename_20240712_110059.png", 
    #                     "zip":   "702/702702_results_20240712_110059.zip"}'
    #              }] 
    #         count=None
    #
    # or if none found:
    #         data=[] count=None
    #
    ########################################################

    
    if len(result.data)==0:
        msg      = f'\n\n\t[pr] ***** NO  status=="COMPLETE"  JOBS WERE FOUND (all have been fetched) ***** \n\n'
        return { 'task_id':  None, 
                 'error':    False,
                 'message':  msg}     
        
    
    jj           =  result.data[0] 
    task_id      =  jj[ 'task_id']                      
    render_value = json.loads( jj['render_value'])      # Convert text representation to a dictionary
    image_path   = render_value['image']                # Extract items from render_value dictionary
    zip_path     = render_value['zip']

    logger.info(f"*** OLDEST COMPLETE *** found:\n\tTask ID:\t{task_id}\n\tImage Path:\t{image_path}\n\tZip Path:\t{zip_path}\n")
        
    return  {'task_id':     task_id,   
             'image_path':  image_path, 
             'zip_path':    zip_path, 
             'error':       False,
             'message':    'oldest complete found ok'}
    
    
    
    
    
    
    

# -------------------------------------------------------------  test/dev code only

    
def  start_fetch_in_background_thread( results_folder):
    
     # fetching runs in its own thread -- with 'Fetch' in its name
    
     thread = threading.Thread( target=fetch_any_unfetched_results, args=( results_folder,) )   # NB that comma is crucial 
                                                                                            # args=( results_folder) is not a tuple, get strange errors
     
     thread.daemon = True     # Set the thread as a daemon so it runs in the background ok/good for plugin, as stays open for long enough
                              # but not good for quick test, as when the main thread exits, the daemon thread (half way through work) will be stopped
                              # this is dealt with in test below by having a sleep n-seconds wait while fetching the COMPLETEs
     
     thread.start()
    
    
    
  